from __future__ import annotations

from collections import defaultdict
from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.cogshop1.data_classes import (
    Scenario,
    ScenarioApply,
    ScenarioApplyList,
    ScenarioFields,
    ScenarioList,
    ScenarioTextFields,
)
from cognite.powerops.client._generated.cogshop1.data_classes._scenario import _SCENARIO_PROPERTIES_BY_FIELD

from ._core import DEFAULT_LIMIT_READ, IN_FILTER_LIMIT, Aggregations, TypeAPI


class ScenarioExtraFilesAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str], space="cogShop") -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": space, "externalId": "Scenario.extraFiles"},
        )
        if isinstance(external_id, str):
            is_scenario = f.Equals(
                ["edge", "startNode"],
                {"space": space, "externalId": external_id},
            )
            return self._client.data_modeling.instances.list("edge", limit=-1, filter=f.And(is_edge_type, is_scenario))

        else:
            is_scenarios = f.In(
                ["edge", "startNode"],
                [{"space": space, "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list("edge", limit=-1, filter=f.And(is_edge_type, is_scenarios))

    def list(
        self, scenario_id: str | list[str] | None = None, limit=DEFAULT_LIMIT_READ, space="cogShop"
    ) -> dm.EdgeList:
        f = dm.filters
        filters = []
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": space, "externalId": "Scenario.extraFiles"},
        )
        filters.append(is_edge_type)
        if scenario_id:
            scenario_ids = [scenario_id] if isinstance(scenario_id, str) else scenario_id
            is_scenarios = f.In(
                ["edge", "startNode"],
                [{"space": space, "externalId": ext_id} for ext_id in scenario_ids],
            )
            filters.append(is_scenarios)

        return self._client.data_modeling.instances.list("edge", limit=limit, filter=f.And(*filters))


class ScenarioMappingsOverrideAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str], space="cogShop") -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": space, "externalId": "Scenario.mappingsOverride"},
        )
        if isinstance(external_id, str):
            is_scenario = f.Equals(
                ["edge", "startNode"],
                {"space": space, "externalId": external_id},
            )
            return self._client.data_modeling.instances.list("edge", limit=-1, filter=f.And(is_edge_type, is_scenario))

        else:
            is_scenarios = f.In(
                ["edge", "startNode"],
                [{"space": space, "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list("edge", limit=-1, filter=f.And(is_edge_type, is_scenarios))

    def list(
        self, scenario_id: str | list[str] | None = None, limit=DEFAULT_LIMIT_READ, space="cogShop"
    ) -> dm.EdgeList:
        f = dm.filters
        filters = []
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": space, "externalId": "Scenario.mappingsOverride"},
        )
        filters.append(is_edge_type)
        if scenario_id:
            scenario_ids = [scenario_id] if isinstance(scenario_id, str) else scenario_id
            is_scenarios = f.In(
                ["edge", "startNode"],
                [{"space": space, "externalId": ext_id} for ext_id in scenario_ids],
            )
            filters.append(is_scenarios)

        return self._client.data_modeling.instances.list("edge", limit=limit, filter=f.And(*filters))


class ScenarioAPI(TypeAPI[Scenario, ScenarioApply, ScenarioList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=Scenario,
            class_apply_type=ScenarioApply,
            class_list=ScenarioList,
        )
        self._view_id = view_id
        self.extra_files = ScenarioExtraFilesAPI(client)
        self.mappings_override = ScenarioMappingsOverrideAPI(client)

    def apply(
        self, scenario: ScenarioApply | Sequence[ScenarioApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(scenario, ScenarioApply):
            instances = scenario.to_instances_apply()
        else:
            instances = ScenarioApplyList(scenario).to_instances_apply()
        return self._client.data_modeling.instances.apply(
            nodes=instances.nodes,
            edges=instances.edges,
            auto_create_start_nodes=True,
            auto_create_end_nodes=True,
            replace=replace,
        )

    def delete(self, external_id: str | Sequence[str], space="cogShop") -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> Scenario: ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> ScenarioList: ...

    def retrieve(self, external_id: str | Sequence[str]) -> Scenario | ScenarioList:
        if isinstance(external_id, str):
            scenario = self._retrieve((self._sources.space, external_id))

            extra_file_edges = self.extra_files.retrieve(external_id)
            scenario.extra_files = [edge.end_node.external_id for edge in extra_file_edges]
            mappings_override_edges = self.mappings_override.retrieve(external_id)
            scenario.mappings_override = [edge.end_node.external_id for edge in mappings_override_edges]

            return scenario
        else:
            scenarios = self._retrieve([(self._sources.space, ext_id) for ext_id in external_id])

            extra_file_edges = self.extra_files.retrieve(external_id)
            self._set_extra_files(scenarios, extra_file_edges)
            mappings_override_edges = self.mappings_override.retrieve(external_id)
            self._set_mappings_override(scenarios, mappings_override_edges)

            return scenarios

    def search(
        self,
        query: str,
        properties: ScenarioTextFields | Sequence[ScenarioTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        model_template: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        commands: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        source: str | list[str] | None = None,
        source_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> ScenarioList:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            model_template,
            commands,
            source,
            source_prefix,
            external_id_prefix,
            filter,
        )
        return self._search(self._view_id, query, _SCENARIO_PROPERTIES_BY_FIELD, properties, filter_, limit)

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: ScenarioFields | Sequence[ScenarioFields] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: ScenarioTextFields | Sequence[ScenarioTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        model_template: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        commands: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        source: str | list[str] | None = None,
        source_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]: ...

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: ScenarioFields | Sequence[ScenarioFields] | None = None,
        group_by: ScenarioFields | Sequence[ScenarioFields] = None,
        query: str | None = None,
        search_properties: ScenarioTextFields | Sequence[ScenarioTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        model_template: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        commands: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        source: str | list[str] | None = None,
        source_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList: ...

    def aggregate(
        self,
        aggregate: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: ScenarioFields | Sequence[ScenarioFields] | None = None,
        group_by: ScenarioFields | Sequence[ScenarioFields] | None = None,
        query: str | None = None,
        search_property: ScenarioTextFields | Sequence[ScenarioTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        model_template: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        commands: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        source: str | list[str] | None = None,
        source_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            model_template,
            commands,
            source,
            source_prefix,
            external_id_prefix,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _SCENARIO_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: ScenarioFields,
        interval: float,
        query: str | None = None,
        search_property: ScenarioTextFields | Sequence[ScenarioTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        model_template: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        commands: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        source: str | list[str] | None = None,
        source_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            model_template,
            commands,
            source,
            source_prefix,
            external_id_prefix,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _SCENARIO_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        model_template: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        commands: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        source: str | list[str] | None = None,
        source_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        retrieve_edges: bool = True,
    ) -> ScenarioList:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            model_template,
            commands,
            source,
            source_prefix,
            external_id_prefix,
            filter,
        )

        scenarios = self._list(limit=limit, filter=filter_)

        if retrieve_edges:
            if len(external_ids := scenarios.as_external_ids()) > IN_FILTER_LIMIT:
                extra_file_edges = self.extra_files.list(limit=-1)
            else:
                extra_file_edges = self.extra_files.list(external_ids, limit=-1)
            self._set_extra_files(scenarios, extra_file_edges)
            if len(external_ids := scenarios.as_external_ids()) > IN_FILTER_LIMIT:
                mappings_override_edges = self.mappings_override.list(limit=-1)
            else:
                mappings_override_edges = self.mappings_override.list(external_ids, limit=-1)
            self._set_mappings_override(scenarios, mappings_override_edges)

        return scenarios

    @staticmethod
    def _set_extra_files(scenarios: Sequence[Scenario], extra_file_edges: Sequence[dm.Edge]):
        edges_by_start_node: dict[tuple, list] = defaultdict(list)
        for edge in extra_file_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for scenario in scenarios:
            node_id = scenario.id_tuple()
            if node_id in edges_by_start_node:
                scenario.extra_files = [edge.end_node.external_id for edge in edges_by_start_node[node_id]]

    @staticmethod
    def _set_mappings_override(scenarios: Sequence[Scenario], mappings_override_edges: Sequence[dm.Edge]):
        edges_by_start_node: dict[tuple, list] = defaultdict(list)
        for edge in mappings_override_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for scenario in scenarios:
            node_id = scenario.id_tuple()
            if node_id in edges_by_start_node:
                scenario.mappings_override = [edge.end_node.external_id for edge in edges_by_start_node[node_id]]


def _create_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    model_template: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    commands: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    source: str | list[str] | None = None,
    source_prefix: str | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if name and isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if model_template and isinstance(model_template, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("modelTemplate"), value={"space": "cogShop", "externalId": model_template}
            )
        )
    if model_template and isinstance(model_template, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("modelTemplate"),
                value={"space": model_template[0], "externalId": model_template[1]},
            )
        )
    if model_template and isinstance(model_template, list) and isinstance(model_template[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("modelTemplate"),
                values=[{"space": "cogShop", "externalId": item} for item in model_template],
            )
        )
    if model_template and isinstance(model_template, list) and isinstance(model_template[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("modelTemplate"),
                values=[{"space": item[0], "externalId": item[1]} for item in model_template],
            )
        )
    if commands and isinstance(commands, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("commands"), value={"space": "cogShop", "externalId": commands})
        )
    if commands and isinstance(commands, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("commands"), value={"space": commands[0], "externalId": commands[1]}
            )
        )
    if commands and isinstance(commands, list) and isinstance(commands[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("commands"),
                values=[{"space": "cogShop", "externalId": item} for item in commands],
            )
        )
    if commands and isinstance(commands, list) and isinstance(commands[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("commands"),
                values=[{"space": item[0], "externalId": item[1]} for item in commands],
            )
        )
    if source and isinstance(source, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("source"), value=source))
    if source and isinstance(source, list):
        filters.append(dm.filters.In(view_id.as_property_ref("source"), values=source))
    if source_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("source"), value=source_prefix))
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import datetime
from collections.abc import Sequence
from typing import Literal

import pandas as pd
from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes import Datapoints, DatapointsArrayList, DatapointsList, TimeSeriesList
from cognite.client.data_classes.datapoints import Aggregate
from cognite.powerops.client._generated.v1.data_classes._plant_water_value_based import (
    _create_plant_water_value_based_filter,
)
from ._core import DEFAULT_LIMIT_READ, INSTANCE_QUERY_LIMIT

ColumnNames = Literal[
    "name",
    "displayName",
    "ordering",
    "assetType",
    "headLossFactor",
    "outletLevel",
    "productionMax",
    "productionMin",
    "connectionLosses",
    "productionMaxTimeSeries",
    "productionMinTimeSeries",
    "waterValueTimeSeries",
    "feedingFeeTimeSeries",
    "outletLevelTimeSeries",
    "inletLevelTimeSeries",
    "headDirectTimeSeries",
]


class PlantWaterValueBasedHeadDirectTimeSeriesQuery:
    def __init__(
        self,
        client: CogniteClient,
        view_id: dm.ViewId,
        timeseries_limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ):
        self._client = client
        self._view_id = view_id
        self._timeseries_limit = timeseries_limit
        self._filter = filter

    def retrieve(
        self,
        start: int | str | datetime.datetime | None = None,
        end: int | str | datetime.datetime | None = None,
        *,
        aggregates: Aggregate | list[Aggregate] | None = None,
        granularity: str | None = None,
        target_unit: str | None = None,
        target_unit_system: str | None = None,
        limit: int | None = None,
        include_outside_points: bool = False,
    ) -> DatapointsList:
        """`Retrieve datapoints for the `plant_water_value_based.head_direct_time_series` timeseries.

        **Performance guide**:
            In order to retrieve millions of datapoints as efficiently as possible, here are a few guidelines:

            1. For the best speed, and significantly lower memory usage, consider using ``retrieve_arrays(...)`` which uses ``numpy.ndarrays`` for data storage.
            2. Only unlimited queries with (``limit=None``) are fetched in parallel, so specifying a large finite ``limit`` like 1 million, comes with severe performance penalty as data is fetched serially.
            3. Try to avoid specifying `start` and `end` to be very far from the actual data: If you had data from 2000 to 2015, don't set start=0 (1970).

        Args:
            start: Inclusive start. Default: 1970-01-01 UTC.
            end: Exclusive end. Default: "now"
            aggregates: Single aggregate or list of aggregates to retrieve. Default: None (raw datapoints returned)
            granularity The granularity to fetch aggregates at. e.g. '15s', '2h', '10d'. Default: None.
            target_unit: The unit_external_id of the data points returned. If the time series does not have an unit_external_id that can be converted to the target_unit, an error will be returned. Cannot be used with target_unit_system.
            target_unit_system: The unit system of the data points returned. Cannot be used with target_unit.
            limit (int | None): Maximum number of datapoints to return for each time series. Default: None (no limit)
            include_outside_points (bool): Whether to include outside points. Not allowed when fetching aggregates. Default: False

        Returns:
            A ``DatapointsList`` with the requested datapoints.

        Examples:

            In this example,
            we are using the time-ago format to get raw data for the 'my_head_direct_time_series' from 2 weeks ago up until now::

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> plant_water_value_based_datapoints = client.plant_water_value_based.head_direct_time_series(external_id="my_head_direct_time_series").retrieve(start="2w-ago")
        """
        external_ids = self._retrieve_timeseries_external_ids_with_extra()
        if external_ids:
            return self._client.time_series.data.retrieve(
                external_id=list(external_ids),
                start=start,
                end=end,
                aggregates=aggregates,
                granularity=granularity,
                target_unit=target_unit,
                target_unit_system=target_unit_system,
                limit=limit,
                include_outside_points=include_outside_points,
            )
        else:
            return DatapointsList([])

    def retrieve_arrays(
        self,
        start: int | str | datetime.datetime | None = None,
        end: int | str | datetime.datetime | None = None,
        *,
        aggregates: Aggregate | list[Aggregate] | None = None,
        granularity: str | None = None,
        target_unit: str | None = None,
        target_unit_system: str | None = None,
        limit: int | None = None,
        include_outside_points: bool = False,
    ) -> DatapointsArrayList:
        """`Retrieve numpy arrays for the `plant_water_value_based.head_direct_time_series` timeseries.

        **Performance guide**:
            In order to retrieve millions of datapoints as efficiently as possible, here are a few guidelines:

            1. For the best speed, and significantly lower memory usage, consider using ``retrieve_arrays(...)`` which uses ``numpy.ndarrays`` for data storage.
            2. Only unlimited queries with (``limit=None``) are fetched in parallel, so specifying a large finite ``limit`` like 1 million, comes with severe performance penalty as data is fetched serially.
            3. Try to avoid specifying `start` and `end` to be very far from the actual data: If you had data from 2000 to 2015, don't set start=0 (1970).

        Args:
            start: Inclusive start. Default: 1970-01-01 UTC.
            end: Exclusive end. Default: "now"
            aggregates: Single aggregate or list of aggregates to retrieve. Default: None (raw datapoints returned)
            granularity The granularity to fetch aggregates at. e.g. '15s', '2h', '10d'. Default: None.
            target_unit: The unit_external_id of the data points returned. If the time series does not have an unit_external_id that can be converted to the target_unit, an error will be returned. Cannot be used with target_unit_system.
            target_unit_system: The unit system of the data points returned. Cannot be used with target_unit.
            limit (int | None): Maximum number of datapoints to return for each time series. Default: None (no limit)
            include_outside_points (bool): Whether to include outside points. Not allowed when fetching aggregates. Default: False

        Returns:
            A ``DatapointsArrayList`` with the requested datapoints.

        Examples:

            In this example,
            we are using the time-ago format to get raw data for the 'my_head_direct_time_series' from 2 weeks ago up until now::

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> plant_water_value_based_datapoints = client.plant_water_value_based.head_direct_time_series(external_id="my_head_direct_time_series").retrieve_array(start="2w-ago")
        """
        external_ids = self._retrieve_timeseries_external_ids_with_extra()
        if external_ids:
            return self._client.time_series.data.retrieve_arrays(
                external_id=list(external_ids),
                start=start,
                end=end,
                aggregates=aggregates,
                granularity=granularity,
                target_unit=target_unit,
                target_unit_system=target_unit_system,
                limit=limit,
                include_outside_points=include_outside_points,
            )
        else:
            return DatapointsArrayList([])

    def retrieve_dataframe(
        self,
        start: int | str | datetime.datetime | None = None,
        end: int | str | datetime.datetime | None = None,
        *,
        aggregates: Aggregate | list[Aggregate] | None = None,
        granularity: str | None = None,
        target_unit: str | None = None,
        target_unit_system: str | None = None,
        limit: int | None = None,
        include_outside_points: bool = False,
        uniform_index: bool = False,
        include_aggregate_name: bool = True,
        include_granularity_name: bool = False,
        column_names: ColumnNames | list[ColumnNames] = "headDirectTimeSeries",
    ) -> pd.DataFrame:
        """`Retrieve DataFrames for the `plant_water_value_based.head_direct_time_series` timeseries.

        **Performance guide**:
            In order to retrieve millions of datapoints as efficiently as possible, here are a few guidelines:

            1. For the best speed, and significantly lower memory usage, consider using ``retrieve_arrays(...)`` which uses ``numpy.ndarrays`` for data storage.
            2. Only unlimited queries with (``limit=None``) are fetched in parallel, so specifying a large finite ``limit`` like 1 million, comes with severe performance penalty as data is fetched serially.
            3. Try to avoid specifying `start` and `end` to be very far from the actual data: If you had data from 2000 to 2015, don't set start=0 (1970).

        Args:
            start: Inclusive start. Default: 1970-01-01 UTC.
            end: Exclusive end. Default: "now"
            aggregates: Single aggregate or list of aggregates to retrieve. Default: None (raw datapoints returned)
            granularity The granularity to fetch aggregates at. e.g. '15s', '2h', '10d'. Default: None.
            target_unit: The unit_external_id of the data points returned. If the time series does not have an unit_external_id that can be converted to the target_unit, an error will be returned. Cannot be used with target_unit_system.
            target_unit_system: The unit system of the data points returned. Cannot be used with target_unit.
            limit: Maximum number of datapoints to return for each time series. Default: None (no limit)
            include_outside_points: Whether to include outside points. Not allowed when fetching aggregates. Default: False
            uniform_index: If only querying aggregates AND a single granularity is used, AND no limit is used, specifying `uniform_index=True` will return a dataframe with an equidistant datetime index from the earliest `start` to the latest `end` (missing values will be NaNs). If these requirements are not met, a ValueError is raised. Default: False
            include_aggregate_name: Include 'aggregate' in the column name, e.g. `my-ts|average`. Ignored for raw time series. Default: True
            include_granularity_name: Include 'granularity' in the column name, e.g. `my-ts|12h`. Added after 'aggregate' when present. Ignored for raw time series. Default: False
            column_names: Which property to use for column names. Defauts to headDirectTimeSeries


        Returns:
            A ``DataFrame`` with the requested datapoints.

        Examples:

            In this example,
            we are using the time-ago format to get raw data for the 'my_head_direct_time_series' from 2 weeks ago up until now::

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> plant_water_value_based_datapoints = client.plant_water_value_based.head_direct_time_series(external_id="my_head_direct_time_series").retrieve_dataframe(start="2w-ago")
        """
        external_ids = self._retrieve_timeseries_external_ids_with_extra(column_names)
        if external_ids:
            df = self._client.time_series.data.retrieve_dataframe(
                external_id=list(external_ids),
                start=start,
                end=end,
                aggregates=aggregates,
                granularity=granularity,
                target_unit=target_unit,
                target_unit_system=target_unit_system,
                limit=limit,
                include_outside_points=include_outside_points,
                uniform_index=uniform_index,
                include_aggregate_name=include_aggregate_name,
                include_granularity_name=include_granularity_name,
            )
            is_aggregate = aggregates is not None
            return self._rename_columns(
                external_ids,
                df,
                column_names,
                is_aggregate and include_aggregate_name,
                is_aggregate and include_granularity_name,
            )
        else:
            return pd.DataFrame()

    def retrieve_dataframe_in_tz(
        self,
        start: datetime.datetime,
        end: datetime.datetime,
        *,
        aggregates: Aggregate | Sequence[Aggregate] | None = None,
        granularity: str | None = None,
        target_unit: str | None = None,
        target_unit_system: str | None = None,
        uniform_index: bool = False,
        include_aggregate_name: bool = True,
        include_granularity_name: bool = False,
        column_names: ColumnNames | list[ColumnNames] = "headDirectTimeSeries",
    ) -> pd.DataFrame:
        """Retrieve DataFrames for the `plant_water_value_based.head_direct_time_series` timeseries in Timezone.

        **Performance guide**:
            In order to retrieve millions of datapoints as efficiently as possible, here are a few guidelines:

            1. For the best speed, and significantly lower memory usage, consider using ``retrieve_arrays(...)`` which uses ``numpy.ndarrays`` for data storage.
            2. Only unlimited queries with (``limit=None``) are fetched in parallel, so specifying a large finite ``limit`` like 1 million, comes with severe performance penalty as data is fetched serially.
            3. Try to avoid specifying `start` and `end` to be very far from the actual data: If you had data from 2000 to 2015, don't set start=0 (1970).

        Args:
            start: Inclusive start.
            end: Exclusive end
            aggregates: Single aggregate or list of aggregates to retrieve. Default: None (raw datapoints returned)
            granularity The granularity to fetch aggregates at. e.g. '15s', '2h', '10d'. Default: None.
            target_unit: The unit_external_id of the data points returned. If the time series does not have an unit_external_id that can be converted to the target_unit, an error will be returned. Cannot be used with target_unit_system.
            target_unit_system: The unit system of the data points returned. Cannot be used with target_unit.
            limit: Maximum number of datapoints to return for each time series. Default: None (no limit)
            include_outside_points: Whether to include outside points. Not allowed when fetching aggregates. Default: False
            uniform_index: If only querying aggregates AND a single granularity is used, AND no limit is used, specifying `uniform_index=True` will return a dataframe with an equidistant datetime index from the earliest `start` to the latest `end` (missing values will be NaNs). If these requirements are not met, a ValueError is raised. Default: False
            include_aggregate_name: Include 'aggregate' in the column name, e.g. `my-ts|average`. Ignored for raw time series. Default: True
            include_granularity_name: Include 'granularity' in the column name, e.g. `my-ts|12h`. Added after 'aggregate' when present. Ignored for raw time series. Default: False
            column_names: Which property to use for column names. Defauts to headDirectTimeSeries


        Returns:
            A ``DataFrame`` with the requested datapoints.

        Examples:

            In this example,
            get weekly aggregates for the 'my_head_direct_time_series' for the first month of 2023 in Oslo time:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> from datetime import datetime, timezone
                >>> client = PowerOpsModelsV1Client()
                >>> plant_water_value_based_datapoints = client.plant_water_value_based.head_direct_time_series(
                ...     external_id="my_head_direct_time_series").retrieve_dataframe_in_timezone(
                ...         datetime(2023, 1, 1, tzinfo=ZoneInfo("Europe/Oslo")),
                ...         datetime(2023, 1, 2, tzinfo=ZoneInfo("Europe/Oslo")),
                ...         aggregates="average",
                ...         granularity="1week",
                ...     )
        """
        external_ids = self._retrieve_timeseries_external_ids_with_extra(column_names)
        if external_ids:
            df = self._client.time_series.data.retrieve_dataframe_in_tz(
                external_id=list(external_ids),
                start=start,
                end=end,
                aggregates=aggregates,
                granularity=granularity,
                target_unit=target_unit,
                target_unit_system=target_unit_system,
                uniform_index=uniform_index,
                include_aggregate_name=include_aggregate_name,
                include_granularity_name=include_granularity_name,
            )
            is_aggregate = aggregates is not None
            return self._rename_columns(
                external_ids,
                df,
                column_names,
                is_aggregate and include_aggregate_name,
                is_aggregate and include_granularity_name,
            )
        else:
            return pd.DataFrame()

    def retrieve_latest(
        self,
        before: None | int | str | datetime.datetime = None,
    ) -> Datapoints | DatapointsList | None:
        external_ids = self._retrieve_timeseries_external_ids_with_extra()
        if external_ids:
            return self._client.time_series.data.retrieve_latest(
                external_id=list(external_ids),
                before=before,
            )
        else:
            return None

    def _retrieve_timeseries_external_ids_with_extra(
        self, extra_properties: ColumnNames | list[ColumnNames] = "headDirectTimeSeries"
    ) -> dict[str, list[str]]:
        return _retrieve_timeseries_external_ids_with_extra_head_direct_time_series(
            self._client,
            self._view_id,
            self._filter,
            self._timeseries_limit,
            extra_properties,
        )

    @staticmethod
    def _rename_columns(
        external_ids: dict[str, list[str]],
        df: pd.DataFrame,
        column_names: ColumnNames | list[ColumnNames],
        include_aggregate_name: bool,
        include_granularity_name: bool,
    ) -> pd.DataFrame:
        if isinstance(column_names, str) and column_names == "headDirectTimeSeries":
            return df
        splits = sum(included for included in [include_aggregate_name, include_granularity_name])
        if splits == 0:
            df.columns = ["-".join(external_ids[external_id]) for external_id in df.columns]
        else:
            column_parts = (col.rsplit("|", maxsplit=splits) for col in df.columns)
            df.columns = [
                "-".join(external_ids[external_id]) + "|" + "|".join(parts) for external_id, *parts in column_parts
            ]
        return df


class PlantWaterValueBasedHeadDirectTimeSeriesAPI:
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        self._client = client
        self._view_id = view_id

    def __call__(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        asset_type: str | list[str] | None = None,
        asset_type_prefix: str | None = None,
        min_head_loss_factor: float | None = None,
        max_head_loss_factor: float | None = None,
        min_outlet_level: float | None = None,
        max_outlet_level: float | None = None,
        min_production_max: float | None = None,
        max_production_max: float | None = None,
        min_production_min: float | None = None,
        max_production_min: float | None = None,
        min_connection_losses: float | None = None,
        max_connection_losses: float | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> PlantWaterValueBasedHeadDirectTimeSeriesQuery:
        """Query timeseries `plant_water_value_based.head_direct_time_series`

        Args:
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            display_name: The display name to filter on.
            display_name_prefix: The prefix of the display name to filter on.
            min_ordering: The minimum value of the ordering to filter on.
            max_ordering: The maximum value of the ordering to filter on.
            asset_type: The asset type to filter on.
            asset_type_prefix: The prefix of the asset type to filter on.
            min_head_loss_factor: The minimum value of the head loss factor to filter on.
            max_head_loss_factor: The maximum value of the head loss factor to filter on.
            min_outlet_level: The minimum value of the outlet level to filter on.
            max_outlet_level: The maximum value of the outlet level to filter on.
            min_production_max: The minimum value of the production max to filter on.
            max_production_max: The maximum value of the production max to filter on.
            min_production_min: The minimum value of the production min to filter on.
            max_production_min: The maximum value of the production min to filter on.
            min_connection_losses: The minimum value of the connection loss to filter on.
            max_connection_losses: The maximum value of the connection loss to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of plant water value baseds to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            A query object that can be used to retrieve datapoins for the plant_water_value_based.head_direct_time_series timeseries
            selected in this method.

        Examples:

            Retrieve all data for 5 plant_water_value_based.head_direct_time_series timeseries:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> plant_water_value_baseds = client.plant_water_value_based.head_direct_time_series(limit=5).retrieve()

        """
        filter_ = _create_plant_water_value_based_filter(
            self._view_id,
            name,
            name_prefix,
            display_name,
            display_name_prefix,
            min_ordering,
            max_ordering,
            asset_type,
            asset_type_prefix,
            min_head_loss_factor,
            max_head_loss_factor,
            min_outlet_level,
            max_outlet_level,
            min_production_max,
            max_production_max,
            min_production_min,
            max_production_min,
            min_connection_losses,
            max_connection_losses,
            external_id_prefix,
            space,
            filter,
        )

        return PlantWaterValueBasedHeadDirectTimeSeriesQuery(
            client=self._client,
            view_id=self._view_id,
            timeseries_limit=limit,
            filter=filter_,
        )

    def list(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        asset_type: str | list[str] | None = None,
        asset_type_prefix: str | None = None,
        min_head_loss_factor: float | None = None,
        max_head_loss_factor: float | None = None,
        min_outlet_level: float | None = None,
        max_outlet_level: float | None = None,
        min_production_max: float | None = None,
        max_production_max: float | None = None,
        min_production_min: float | None = None,
        max_production_min: float | None = None,
        min_connection_losses: float | None = None,
        max_connection_losses: float | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> TimeSeriesList:
        """List timeseries `plant_water_value_based.head_direct_time_series`

        Args:
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            display_name: The display name to filter on.
            display_name_prefix: The prefix of the display name to filter on.
            min_ordering: The minimum value of the ordering to filter on.
            max_ordering: The maximum value of the ordering to filter on.
            asset_type: The asset type to filter on.
            asset_type_prefix: The prefix of the asset type to filter on.
            min_head_loss_factor: The minimum value of the head loss factor to filter on.
            max_head_loss_factor: The maximum value of the head loss factor to filter on.
            min_outlet_level: The minimum value of the outlet level to filter on.
            max_outlet_level: The maximum value of the outlet level to filter on.
            min_production_max: The minimum value of the production max to filter on.
            max_production_max: The maximum value of the production max to filter on.
            min_production_min: The minimum value of the production min to filter on.
            max_production_min: The maximum value of the production min to filter on.
            min_connection_losses: The minimum value of the connection loss to filter on.
            max_connection_losses: The maximum value of the connection loss to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of plant water value baseds to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            List of Timeseries plant_water_value_based.head_direct_time_series.

        Examples:

            List plant_water_value_based.head_direct_time_series and limit to 5:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> plant_water_value_baseds = client.plant_water_value_based.head_direct_time_series.list(limit=5)

        """
        filter_ = _create_plant_water_value_based_filter(
            self._view_id,
            name,
            name_prefix,
            display_name,
            display_name_prefix,
            min_ordering,
            max_ordering,
            asset_type,
            asset_type_prefix,
            min_head_loss_factor,
            max_head_loss_factor,
            min_outlet_level,
            max_outlet_level,
            min_production_max,
            max_production_max,
            min_production_min,
            max_production_min,
            min_connection_losses,
            max_connection_losses,
            external_id_prefix,
            space,
            filter,
        )
        external_ids = _retrieve_timeseries_external_ids_with_extra_head_direct_time_series(
            self._client, self._view_id, filter_, limit
        )
        if external_ids:
            return self._client.time_series.retrieve_multiple(external_ids=list(external_ids))
        else:
            return TimeSeriesList([])


def _retrieve_timeseries_external_ids_with_extra_head_direct_time_series(
    client: CogniteClient,
    view_id: dm.ViewId,
    filter_: dm.Filter | None,
    limit: int,
    extra_properties: ColumnNames | list[ColumnNames] = "headDirectTimeSeries",
) -> dict[str, list[str]]:
    limit = float("inf") if limit is None or limit == -1 else limit
    properties = ["headDirectTimeSeries"]
    if extra_properties == "headDirectTimeSeries":
        ...
    elif isinstance(extra_properties, str) and extra_properties != "headDirectTimeSeries":
        properties.append(extra_properties)
    elif isinstance(extra_properties, list):
        properties.extend([prop for prop in extra_properties if prop != "headDirectTimeSeries"])
    else:
        raise ValueError(f"Invalid value for extra_properties: {extra_properties}")

    if isinstance(extra_properties, str):
        extra_list = [extra_properties]
    else:
        extra_list = extra_properties
    has_data = dm.filters.HasData(views=[view_id])
    has_property = dm.filters.Exists(property=view_id.as_property_ref("headDirectTimeSeries"))
    filter_ = dm.filters.And(filter_, has_data, has_property) if filter_ else dm.filters.And(has_data, has_property)

    cursor = None
    external_ids: dict[str, list[str]] = {}
    total_retrieved = 0
    while True:
        query_limit = max(min(INSTANCE_QUERY_LIMIT, limit - total_retrieved), 0)
        selected_nodes = dm.query.NodeResultSetExpression(filter=filter_, limit=query_limit)
        query = dm.query.Query(
            with_={
                "nodes": selected_nodes,
            },
            select={
                "nodes": dm.query.Select(
                    [dm.query.SourceSelector(view_id, properties)],
                )
            },
            cursors={"nodes": cursor},
        )
        result = client.data_modeling.instances.query(query)
        batch_external_ids = {
            node.properties[view_id]["headDirectTimeSeries"]: [
                node.properties[view_id].get(prop, "") for prop in extra_list
            ]
            for node in result.data["nodes"].data
        }
        total_retrieved += len(batch_external_ids)
        external_ids.update(batch_external_ids)
        cursor = result.cursors["nodes"]
        if total_retrieved >= limit or cursor is None:
            break
    return external_ids



"""
The module contains functions for transforming the ReSync configuration into Cognite Data Model Types which are used
in multiple models. Note the use of the `make_ext_id` function which is used to generate unique external IDs for
all the conversions. This is done to ensure that the same external ID is used for the same object across multiple
transformations.
"""

from __future__ import annotations

import json
from hashlib import md5
from typing import Any


def make_ext_id(arg: Any, class_: type) -> str:
    hash_value = md5()
    if isinstance(arg, (str, int, float, bool)):
        hash_value.update(str(arg).encode())
    elif isinstance(arg, (list, dict, tuple)):
        hash_value.update(json.dumps(arg).encode())
    return f"{class_.__name__.removesuffix('Apply')}__{hash_value.hexdigest()}"



"""
Checks whether time series referenced in a model actually exist in CDF
and have data in the required time ranges.
"""

from __future__ import annotations

#
#
#
#
#
# class ValidationSpec(BaseModel):
#     pass

#
# class TimeSeriesValidation(ValidationSpec):
#
#
#
# class ValidationRange(BaseModel):
#
#
#     def __str__(self):
#
#     def __eq__(self, other: object) -> bool:
#         if isinstance(other, ValidationRange):
#
#     @field_serializer("start", "end")
#     def serialize_arrow(self, value: arrow.Arrow, _info):
#
#
# class TimeSeriesValidationFailures(str, Enum):
#
#
# class ValidationResult(BaseModel):
#
#
# class ValidationResults(BaseModel):
#
#     def as_markdown(self, include_valid: bool) -> str:
#
#     def as_json(self, include_valid: bool) -> str:
#         if not include_valid:
#
#
# class TimeSeriesValidationResults(ValidationResults):
#
#     def as_markdown(self, include_valid: bool) -> str:
#             "| Time Range                                            | Valid | models | reason | ExternalID |",
#             "|-------------------------------------------------------|-------|--------|--------|------------|",
#         for result in self.results:
#             if result.valid and not include_valid:
#
#
# class TimeSeriesValidationResult(ValidationResult):
#
#     @model_validator(mode="before")
#     @classmethod
#     def set_valid(cls, data: dict) -> dict:
#         if "valid" not in data and "failure" not in data:
#         if data.get("valid") and "failure" in data:
#         if "valid" not in data:
#
#     @property
#     def reason(self) -> str:
#
#
# def find_mappings(obj, lookup: Literal["base", "process"]) -> list[InputTimeSeriesMappingApply]:
#     def df_to_dict_records(item, attr_path):
#
#     def to_input_time_series_mapping_apply(item, _attr_path):
#         if item.get("time_series_external_id") is not None:
#             yield InputTimeSeriesMappingApply(
#
#     def from_mapping_to_input_time_series_mapping_apply(item, _attr_path):
#         if item.timeseries_external_id is not None:
#             yield InputTimeSeriesMappingApply(
#
#         "process": [
#         ],
#         "base": [
#             #  ^ includes scenario mappings
#         ],
#
#     for lookup_path in lookup_paths[lookup]:
#
#
#
#
# def prepare_validation(models: list[Model]) -> tuple[PreparedValidationsT, ValidationRangesT]:
#     # data containers:
#
#     # Find processes (i.e. configuration for actual SHOP runs, with start and end times)
#     # Without this we don't know what time ranges to query from CDF.
#     for model in models:
#         if not hasattr(model, "processes"):
#         for process in model.processes:
#
#             # Find mappings that are configured within processes.
#             # For these mapping we only validate time ranges of the process.
#             for mapping in process_mappings:
#                 if ts_validation is None:
#                 if model.model_name not in ts_validation.data_models:
#
#     # Find base mappings.
#     # For these mappings we don't know the time ranges, so we validate all time ranges.
#     # Overwrite any process-mappings from above.
#     for model in models:
#         for mapping in base_mappings:
#             for validation_range_str, validation_range in validation_ranges.items():
#                 if model.model_name not in ts_validation.data_models:
#
#     # Get any additional timeseries that are part of the models but are not part of the mappings.
#     # These are ambiguous... TODO Overwrite previous mappings or not? These are hard to connect back to SHOP path.
#     for additional_ts in additional_timeseries:
#         for validation_range_str in validation_ranges:
#             ts_validations[validation_range_str][require(additional_ts.external_id)] = TimeSeriesValidation(
#                 mapping=InputTimeSeriesMappingApply(  # dummy mapping, for validation only
#                 ),
#
#
#
# def perform_validation(
#     po_client: PowerOpsClient, ts_validations: PreparedValidationsT, validation_ranges: ValidationRangesT
# ) -> ValidationResults:
#     for range_str, validations_in_range in ts_validations.items():
#
#         for ts_mapping in ts_mappings:
#             if not len(data):
#                 logger.info(
#                 results.append(
#                     TimeSeriesValidationResult(
#                 logger.info(
#                 results.append(
#                     TimeSeriesValidationResult(
#         if warnings_n:



from __future__ import annotations

from functools import cached_property
from pathlib import Path
from typing import Any, ClassVar, Optional, cast

from pydantic import BaseModel, ConfigDict, Field

from cognite.powerops.resync.config._shared import TimeSeriesMapping
from cognite.powerops.utils.serialization import load_yaml

from .generator import Generator


class Watercourse(BaseModel):
    name: str
    shop_penalty_limit: int = 42000


class WatercourseConfig(Watercourse):
    """
    Represents the configuration for a Watercourse

    Attributes:
        version: The version of the watercourse configuration.

    """

    model_config: ClassVar[ConfigDict] = ConfigDict(protected_namespaces=tuple())
    version: str
    market_to_price_area: dict[str, str]
    directory: str
    model_raw: str
    # ------------------------------------------------------------------
    shop_model_template: dict[str, Any]
    yaml_raw_path: Path
    yaml_processed_path: Path  # TODO: not used here
    yaml_mapping_path: str = ""
    model_processed: str  # TODO: not used here
    model_mapping: Optional[str] = None
    write_back_model_file: Optional[bool] = True
    tco_paths: Optional[list[str]] = None  # TODO: not used here - HEV specific
    rrs_ids: Optional[list[str]] = None
    hardcoded_mapping: Optional[TimeSeriesMapping] = None  # TODO: not used here
    hist_flow_timeseries: Optional[dict[str, str]] = None  # TODO: not used here
    # ------------------------------------------------------------------
    production_obligation_ts_ext_ids: list[str] = Field(default_factory=list)
    plant_display_names_and_order: dict[str, tuple[str, int]] = Field(default_factory=dict)
    reservoir_display_names_and_order: dict[str, tuple[str, int]] = Field(default_factory=dict)
    water_value_based_method_time_series_csv_filename: Optional[str] = None
    generators: list[Generator] = Field(default_factory=list)

    @classmethod
    def from_yaml(cls, yaml_path: Path) -> list[WatercourseConfig]:
        return [cls(**watercourse_raw) for watercourse_raw in load_yaml(yaml_path)]

    @cached_property
    def valid_shop_objects(self) -> set[tuple[str, str]]:
        if "model" not in self.shop_model_template:
            raise ValueError("Invalid SHOP yaml: Missing 'model' key")
        model = cast(dict[str, Any], self.shop_model_template["model"])
        valid_mappings: set[tuple[str, str]] = set()
        for object_type, objects in model.items():
            if not isinstance(objects, dict):
                raise ValueError("Invalid SHOP yaml")
            for object_name, attributes in objects.items():
                if not isinstance(attributes, dict):
                    raise ValueError("Invalid SHOP yaml")
                valid_mappings.add((str(object_type).lower(), str(object_name).lower()))
        return valid_mappings



def attr_lookup(value, attr_path):
    """
    Retrieve attributes from a nested structure, in a generator.

    Usage
    Some nested containers:
    >>> class A:
    ...     def __init__(self, b):
    ...         self.b = b
    >>> class B:
    ...     def __init__(self, list_of_cs):
    ...         self.cs = list_of_cs
    >>> class C:
    ...     def __init__(self, x):
    ...         self.x = x
    >>> a = A(B([C(1), C(2)]))

    >>> list(attr_lookup(a, ["b", "cs", each, "x"]))
    [1, 2]

    >>> list(attr_lookup(a, []))  # doctest: +ELLIPSIS
    [<powerops.utils.lookup.A object at 0x...>]
    >>> list(attr_lookup(a, ["b"]))  # doctest: +ELLIPSIS
    [<powerops.utils.lookup.B object at 0x...>]
    >>> list(attr_lookup(a, ["b", "cs"]))  # doctest: +ELLIPSIS
    [[<powerops.utils.lookup.C object at 0x...>, <powerops.utils.lookup.C object at 0x...>]]
    >>> list(attr_lookup(a, ["b", "cs", each]))  # doctest: +ELLIPSIS
    [<powerops.utils.lookup.C object at 0x...>, <powerops.utils.lookup.C object at 0x...>]

    Dicts are supported via special getter:
    >>> a.foo = {"bar": 11, "baz": 22}
    >>> list(attr_lookup(a, ["foo", dict_get, "baz"]))
    [22]

    Errors:
    >>> a.foo = [{"bar": 11}, "blah", {"bar": 33}]
    >>> list(attr_lookup(a, ["foo", each, dict_get, "bar"]))
    [11, None, 33]

    Callables can transform results:
    >>> def stringify(value, _attr_path):
    ...     yield f"here be {value}"
    >>> list(attr_lookup(a, ["foo", each, dict_get, "bar", stringify]))
    ['here be 11', None, 'here be 33']

    Callables can affect the traversal:
    >>> def only_dicts(value, attr_path):
    ...     filtered_value = [val for val in value if isinstance(val, dict)]
    ...     yield from attr_lookup(filtered_value, attr_path)
    >>> list(attr_lookup(a, ["foo", only_dicts, each, dict_get, "bar", stringify]))
    ['here be 11', 'here be 33']
    """
    if len(attr_path) == 0 or value is None:
        yield value
    else:
        sub_attr = attr_path[0]
        if callable(sub_attr):
            yield from sub_attr(value, attr_path[1:])
        else:
            yield from attr_lookup(getattr(value, sub_attr, None), attr_path[1:])


def each(items, attr_path):
    for item in items:
        yield from attr_lookup(item, attr_path)


def dict_get(value, attr_path):
    if isinstance(value, dict):
        yield from attr_lookup(value.get(attr_path[0]), attr_path[1:])
    else:
        yield None


def dict_values(value, attr_path):
    yield from attr_lookup(value.values(), attr_path)



from __future__ import annotations

from typing import Literal, Optional

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

__all__ = ["FileRef", "FileRefApply", "FileRefList", "FileRefApplyList", "FileRefFields", "FileRefTextFields"]


FileRefTextFields = Literal["type", "file_external_id"]
FileRefFields = Literal["type", "file_external_id"]

_FILEREF_PROPERTIES_BY_FIELD = {
    "type": "type",
    "file_external_id": "fileExternalId",
}


class FileRef(DomainModel):
    space: str = "cogShop"
    type: Optional[str] = None
    file_external_id: Optional[str] = Field(None, alias="fileExternalId")

    def as_apply(self) -> FileRefApply:
        return FileRefApply(
            external_id=self.external_id,
            type=self.type,
            file_external_id=self.file_external_id,
        )


class FileRefApply(DomainModelApply):
    space: str = "cogShop"
    type: str
    file_external_id: str = Field(alias="fileExternalId")

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.type is not None:
            properties["type"] = self.type
        if self.file_external_id is not None:
            properties["fileExternalId"] = self.file_external_id
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("cogShop", "FileRef"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class FileRefList(TypeList[FileRef]):
    _NODE = FileRef

    def as_apply(self) -> FileRefApplyList:
        return FileRefApplyList([node.as_apply() for node in self.data])


class FileRefApplyList(TypeApplyList[FileRefApply]):
    _NODE = FileRefApply



from dataclasses import dataclass, field
from pathlib import Path
from typing import Union

import yaml
from cognite.client.data_classes.data_modeling import (
    ContainerApplyList,
    DataModelApply,
    DataModelId,
    ViewApplyList,
)


@dataclass
class PowerOpsDMSSourceModel:
    container_file: Path
    _containers: Union[ContainerApplyList, None] = field(init=False, default=None)

    @property
    def containers(self) -> ContainerApplyList:
        if self._containers is None:
            self._containers = self._load_containers()
        return self._containers

    def _load_containers(self) -> ContainerApplyList:
        return ContainerApplyList._load(yaml.safe_load(self.container_file.read_text()))

    def spaces(self) -> list[str]:
        return list({container.space for container in self.containers})


@dataclass
class PowerOpsDMSModel:
    name: str
    description: str
    id_: DataModelId
    view_file: Path
    _views: Union[ViewApplyList, None] = field(init=False, default=None)

    @property
    def views(self) -> ViewApplyList:
        if self._views is None:
            self._views = self._load_views()
        return self._views

    def _load_views(self) -> ViewApplyList:
        views = ViewApplyList._load(yaml.safe_load(self.view_file.read_text()))
        for view in views:
            view.version = self.id_.version
        return views

    @property
    def data_model(self) -> DataModelApply:
        views = ViewApplyList._load(yaml.safe_load(self.view_file.read_text()))

        return DataModelApply(
            space=self.id_.space,
            external_id=self.id_.external_id,
            version=self.id_.version,
            description=self.description,
            name=self.name,
            views=list(views),
        )

    def spaces(self) -> list[str]:
        return list({view.space for view in self.data_model.views})



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field
from pydantic import field_validator, model_validator

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
)

if TYPE_CHECKING:
    from ._shop_attribute_mapping import ShopAttributeMapping, ShopAttributeMappingGraphQL, ShopAttributeMappingWrite
    from ._shop_commands import ShopCommands, ShopCommandsGraphQL, ShopCommandsWrite
    from ._shop_model import ShopModel, ShopModelGraphQL, ShopModelWrite
    from ._shop_output_time_series_definition import (
        ShopOutputTimeSeriesDefinition,
        ShopOutputTimeSeriesDefinitionGraphQL,
        ShopOutputTimeSeriesDefinitionWrite,
    )


__all__ = [
    "ShopScenario",
    "ShopScenarioWrite",
    "ShopScenarioApply",
    "ShopScenarioList",
    "ShopScenarioWriteList",
    "ShopScenarioApplyList",
    "ShopScenarioFields",
    "ShopScenarioTextFields",
]


ShopScenarioTextFields = Literal["name", "source"]
ShopScenarioFields = Literal["name", "source"]

_SHOPSCENARIO_PROPERTIES_BY_FIELD = {
    "name": "name",
    "source": "source",
}


class ShopScenarioGraphQL(GraphQLCore, protected_namespaces=()):
    """This represents the reading version of shop scenario, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop scenario.
        data_record: The data record of the shop scenario node.
        name: The name of the scenario to run
        model: The model template to use when running the scenario
        commands: The commands to run
        source: The source of the scenario
        output_definition: An array of output definitions for the time series
        attribute_mappings_override: An array of base mappings to override in shop model file
    """

    view_id = dm.ViewId("power_ops_core", "ShopScenario", "1")
    name: Optional[str] = None
    model: Optional[ShopModelGraphQL] = Field(default=None, repr=False)
    commands: Optional[ShopCommandsGraphQL] = Field(default=None, repr=False)
    source: Optional[str] = None
    output_definition: Optional[list[ShopOutputTimeSeriesDefinitionGraphQL]] = Field(
        default=None, repr=False, alias="outputDefinition"
    )
    attribute_mappings_override: Optional[list[ShopAttributeMappingGraphQL]] = Field(
        default=None, repr=False, alias="attributeMappingsOverride"
    )

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values

    @field_validator("model", "commands", "output_definition", "attribute_mappings_override", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    def as_read(self) -> ShopScenario:
        """Convert this GraphQL format of shop scenario to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return ShopScenario(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            name=self.name,
            model=self.model.as_read() if isinstance(self.model, GraphQLCore) else self.model,
            commands=self.commands.as_read() if isinstance(self.commands, GraphQLCore) else self.commands,
            source=self.source,
            output_definition=[output_definition.as_read() for output_definition in self.output_definition or []],
            attribute_mappings_override=[
                attribute_mappings_override.as_read()
                for attribute_mappings_override in self.attribute_mappings_override or []
            ],
        )

    def as_write(self) -> ShopScenarioWrite:
        """Convert this GraphQL format of shop scenario to the writing format."""
        return ShopScenarioWrite(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            name=self.name,
            model=self.model.as_write() if isinstance(self.model, GraphQLCore) else self.model,
            commands=self.commands.as_write() if isinstance(self.commands, GraphQLCore) else self.commands,
            source=self.source,
            output_definition=[output_definition.as_write() for output_definition in self.output_definition or []],
            attribute_mappings_override=[
                attribute_mappings_override.as_write()
                for attribute_mappings_override in self.attribute_mappings_override or []
            ],
        )


class ShopScenario(DomainModel, protected_namespaces=()):
    """This represents the reading version of shop scenario.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop scenario.
        data_record: The data record of the shop scenario node.
        name: The name of the scenario to run
        model: The model template to use when running the scenario
        commands: The commands to run
        source: The source of the scenario
        output_definition: An array of output definitions for the time series
        attribute_mappings_override: An array of base mappings to override in shop model file
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    name: str
    model: Union[ShopModel, str, dm.NodeId, None] = Field(default=None, repr=False)
    commands: Union[ShopCommands, str, dm.NodeId, None] = Field(default=None, repr=False)
    source: Optional[str] = None
    output_definition: Union[list[ShopOutputTimeSeriesDefinition], list[str], list[dm.NodeId], None] = Field(
        default=None, repr=False, alias="outputDefinition"
    )
    attribute_mappings_override: Union[list[ShopAttributeMapping], list[str], list[dm.NodeId], None] = Field(
        default=None, repr=False, alias="attributeMappingsOverride"
    )

    def as_write(self) -> ShopScenarioWrite:
        """Convert this read version of shop scenario to the writing version."""
        return ShopScenarioWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            name=self.name,
            model=self.model.as_write() if isinstance(self.model, DomainModel) else self.model,
            commands=self.commands.as_write() if isinstance(self.commands, DomainModel) else self.commands,
            source=self.source,
            output_definition=[
                output_definition.as_write() if isinstance(output_definition, DomainModel) else output_definition
                for output_definition in self.output_definition or []
            ],
            attribute_mappings_override=[
                (
                    attribute_mappings_override.as_write()
                    if isinstance(attribute_mappings_override, DomainModel)
                    else attribute_mappings_override
                )
                for attribute_mappings_override in self.attribute_mappings_override or []
            ],
        )

    def as_apply(self) -> ShopScenarioWrite:
        """Convert this read version of shop scenario to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class ShopScenarioWrite(DomainModelWrite, protected_namespaces=()):
    """This represents the writing version of shop scenario.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop scenario.
        data_record: The data record of the shop scenario node.
        name: The name of the scenario to run
        model: The model template to use when running the scenario
        commands: The commands to run
        source: The source of the scenario
        output_definition: An array of output definitions for the time series
        attribute_mappings_override: An array of base mappings to override in shop model file
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    name: str
    model: Union[ShopModelWrite, str, dm.NodeId, None] = Field(default=None, repr=False)
    commands: Union[ShopCommandsWrite, str, dm.NodeId, None] = Field(default=None, repr=False)
    source: Optional[str] = None
    output_definition: Union[list[ShopOutputTimeSeriesDefinitionWrite], list[str], list[dm.NodeId], None] = Field(
        default=None, repr=False, alias="outputDefinition"
    )
    attribute_mappings_override: Union[list[ShopAttributeMappingWrite], list[str], list[dm.NodeId], None] = Field(
        default=None, repr=False, alias="attributeMappingsOverride"
    )

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(ShopScenario, dm.ViewId("power_ops_core", "ShopScenario", "1"))

        properties: dict[str, Any] = {}

        if self.name is not None:
            properties["name"] = self.name

        if self.model is not None:
            properties["model"] = {
                "space": self.space if isinstance(self.model, str) else self.model.space,
                "externalId": self.model if isinstance(self.model, str) else self.model.external_id,
            }

        if self.commands is not None:
            properties["commands"] = {
                "space": self.space if isinstance(self.commands, str) else self.commands.space,
                "externalId": self.commands if isinstance(self.commands, str) else self.commands.external_id,
            }

        if self.source is not None or write_none:
            properties["source"] = self.source

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("power_ops_types", "ShopOutputTimeSeriesDefinition")
        for output_definition in self.output_definition or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=output_definition,
                edge_type=edge_type,
                view_by_read_class=view_by_read_class,
                write_none=write_none,
                allow_version_increase=allow_version_increase,
            )
            resources.extend(other_resources)

        edge_type = dm.DirectRelationReference("power_ops_types", "ShopAttributeMapping")
        for attribute_mappings_override in self.attribute_mappings_override or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=attribute_mappings_override,
                edge_type=edge_type,
                view_by_read_class=view_by_read_class,
                write_none=write_none,
                allow_version_increase=allow_version_increase,
            )
            resources.extend(other_resources)

        if isinstance(self.model, DomainModelWrite):
            other_resources = self.model._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.commands, DomainModelWrite):
            other_resources = self.commands._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        return resources


class ShopScenarioApply(ShopScenarioWrite):
    def __new__(cls, *args, **kwargs) -> ShopScenarioApply:
        warnings.warn(
            "ShopScenarioApply is deprecated and will be removed in v1.0. Use ShopScenarioWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "ShopScenario.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class ShopScenarioList(DomainModelList[ShopScenario]):
    """List of shop scenarios in the read version."""

    _INSTANCE = ShopScenario

    def as_write(self) -> ShopScenarioWriteList:
        """Convert these read versions of shop scenario to the writing versions."""
        return ShopScenarioWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> ShopScenarioWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class ShopScenarioWriteList(DomainModelWriteList[ShopScenarioWrite]):
    """List of shop scenarios in the writing version."""

    _INSTANCE = ShopScenarioWrite


class ShopScenarioApplyList(ShopScenarioWriteList): ...


def _create_shop_scenario_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    model: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    commands: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    source: str | list[str] | None = None,
    source_prefix: str | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if model and isinstance(model, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("model"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": model}
            )
        )
    if model and isinstance(model, tuple):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("model"), value={"space": model[0], "externalId": model[1]})
        )
    if model and isinstance(model, list) and isinstance(model[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("model"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in model],
            )
        )
    if model and isinstance(model, list) and isinstance(model[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("model"), values=[{"space": item[0], "externalId": item[1]} for item in model]
            )
        )
    if commands and isinstance(commands, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("commands"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": commands}
            )
        )
    if commands and isinstance(commands, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("commands"), value={"space": commands[0], "externalId": commands[1]}
            )
        )
    if commands and isinstance(commands, list) and isinstance(commands[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("commands"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in commands],
            )
        )
    if commands and isinstance(commands, list) and isinstance(commands[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("commands"),
                values=[{"space": item[0], "externalId": item[1]} for item in commands],
            )
        )
    if isinstance(source, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("source"), value=source))
    if source and isinstance(source, list):
        filters.append(dm.filters.In(view_id.as_property_ref("source"), values=source))
    if source_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("source"), value=source_prefix))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from collections.abc import Sequence
from typing import overload
import warnings

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.v1.data_classes._core import DEFAULT_INSTANCE_SPACE
from cognite.powerops.client._generated.v1.data_classes import (
    DomainModelCore,
    DomainModelWrite,
    ResourcesWriteResult,
    PartialBidMatrixInformationWithScenarios,
    PartialBidMatrixInformationWithScenariosWrite,
    PartialBidMatrixInformationWithScenariosFields,
    PartialBidMatrixInformationWithScenariosList,
    PartialBidMatrixInformationWithScenariosWriteList,
    PartialBidMatrixInformationWithScenariosTextFields,
)
from cognite.powerops.client._generated.v1.data_classes._partial_bid_matrix_information_with_scenarios import (
    _PARTIALBIDMATRIXINFORMATIONWITHSCENARIOS_PROPERTIES_BY_FIELD,
    _create_partial_bid_matrix_information_with_scenario_filter,
)
from ._core import (
    DEFAULT_LIMIT_READ,
    DEFAULT_QUERY_LIMIT,
    Aggregations,
    NodeAPI,
    SequenceNotStr,
    QueryStep,
    QueryBuilder,
)
from .partial_bid_matrix_information_with_scenarios_alerts import PartialBidMatrixInformationWithScenariosAlertsAPI
from .partial_bid_matrix_information_with_scenarios_underlying_bid_matrices import (
    PartialBidMatrixInformationWithScenariosUnderlyingBidMatricesAPI,
)
from .partial_bid_matrix_information_with_scenarios_multi_scenario_input import (
    PartialBidMatrixInformationWithScenariosMultiScenarioInputAPI,
)
from .partial_bid_matrix_information_with_scenarios_query import PartialBidMatrixInformationWithScenariosQueryAPI


class PartialBidMatrixInformationWithScenariosAPI(
    NodeAPI[
        PartialBidMatrixInformationWithScenarios,
        PartialBidMatrixInformationWithScenariosWrite,
        PartialBidMatrixInformationWithScenariosList,
    ]
):
    def __init__(self, client: CogniteClient, view_by_read_class: dict[type[DomainModelCore], dm.ViewId]):
        view_id = view_by_read_class[PartialBidMatrixInformationWithScenarios]
        super().__init__(
            client=client,
            sources=view_id,
            class_type=PartialBidMatrixInformationWithScenarios,
            class_list=PartialBidMatrixInformationWithScenariosList,
            class_write_list=PartialBidMatrixInformationWithScenariosWriteList,
            view_by_read_class=view_by_read_class,
        )
        self._view_id = view_id
        self.alerts_edge = PartialBidMatrixInformationWithScenariosAlertsAPI(client)
        self.underlying_bid_matrices_edge = PartialBidMatrixInformationWithScenariosUnderlyingBidMatricesAPI(client)
        self.multi_scenario_input_edge = PartialBidMatrixInformationWithScenariosMultiScenarioInputAPI(client)

    def __call__(
        self,
        state: str | list[str] | None = None,
        state_prefix: str | None = None,
        power_asset: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_resource_cost: float | None = None,
        max_resource_cost: float | None = None,
        partial_bid_configuration: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_QUERY_LIMIT,
        filter: dm.Filter | None = None,
    ) -> PartialBidMatrixInformationWithScenariosQueryAPI[PartialBidMatrixInformationWithScenariosList]:
        """Query starting at partial bid matrix information with scenarios.

        Args:
            state: The state to filter on.
            state_prefix: The prefix of the state to filter on.
            power_asset: The power asset to filter on.
            min_resource_cost: The minimum value of the resource cost to filter on.
            max_resource_cost: The maximum value of the resource cost to filter on.
            partial_bid_configuration: The partial bid configuration to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of partial bid matrix information with scenarios to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            A query API for partial bid matrix information with scenarios.

        """
        has_data = dm.filters.HasData(views=[self._view_id])
        filter_ = _create_partial_bid_matrix_information_with_scenario_filter(
            self._view_id,
            state,
            state_prefix,
            power_asset,
            min_resource_cost,
            max_resource_cost,
            partial_bid_configuration,
            external_id_prefix,
            space,
            (filter and dm.filters.And(filter, has_data)) or has_data,
        )
        builder = QueryBuilder(PartialBidMatrixInformationWithScenariosList)
        return PartialBidMatrixInformationWithScenariosQueryAPI(
            self._client, builder, self._view_by_read_class, filter_, limit
        )

    def apply(
        self,
        partial_bid_matrix_information_with_scenario: (
            PartialBidMatrixInformationWithScenariosWrite | Sequence[PartialBidMatrixInformationWithScenariosWrite]
        ),
        replace: bool = False,
        write_none: bool = False,
    ) -> ResourcesWriteResult:
        """Add or update (upsert) partial bid matrix information with scenarios.

        Note: This method iterates through all nodes and timeseries linked to partial_bid_matrix_information_with_scenario and creates them including the edges
        between the nodes. For example, if any of `alerts`, `underlying_bid_matrices` or `multi_scenario_input` are set, then these
        nodes as well as any nodes linked to them, and all the edges linking these nodes will be created.

        Args:
            partial_bid_matrix_information_with_scenario: Partial bid matrix information with scenario or sequence of partial bid matrix information with scenarios to upsert.
            replace (bool): How do we behave when a property value exists? Do we replace all matching and existing values with the supplied values (true)?
                Or should we merge in new values for properties together with the existing values (false)? Note: This setting applies for all nodes or edges specified in the ingestion call.
            write_none (bool): This method, will by default, skip properties that are set to None. However, if you want to set properties to None,
                you can set this parameter to True. Note this only applies to properties that are nullable.
        Returns:
            Created instance(s), i.e., nodes, edges, and time series.

        Examples:

            Create a new partial_bid_matrix_information_with_scenario:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> from cognite.powerops.client._generated.v1.data_classes import PartialBidMatrixInformationWithScenariosWrite
                >>> client = PowerOpsModelsV1Client()
                >>> partial_bid_matrix_information_with_scenario = PartialBidMatrixInformationWithScenariosWrite(external_id="my_partial_bid_matrix_information_with_scenario", ...)
                >>> result = client.partial_bid_matrix_information_with_scenarios.apply(partial_bid_matrix_information_with_scenario)

        """
        warnings.warn(
            "The .apply method is deprecated and will be removed in v1.0. "
            "Please use the .upsert method on the client instead. This means instead of "
            "`my_client.partial_bid_matrix_information_with_scenarios.apply(my_items)` please use `my_client.upsert(my_items)`."
            "The motivation is that all apply methods are the same, and having one apply method per API "
            " class encourages users to create items in small batches, which is inefficient."
            "In addition, .upsert method is more descriptive of what the method does.",
            UserWarning,
            stacklevel=2,
        )
        return self._apply(partial_bid_matrix_information_with_scenario, replace, write_none)

    def delete(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> dm.InstancesDeleteResult:
        """Delete one or more partial bid matrix information with scenario.

        Args:
            external_id: External id of the partial bid matrix information with scenario to delete.
            space: The space where all the partial bid matrix information with scenario are located.

        Returns:
            The instance(s), i.e., nodes and edges which has been deleted. Empty list if nothing was deleted.

        Examples:

            Delete partial_bid_matrix_information_with_scenario by id:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> client.partial_bid_matrix_information_with_scenarios.delete("my_partial_bid_matrix_information_with_scenario")
        """
        warnings.warn(
            "The .delete method is deprecated and will be removed in v1.0. "
            "Please use the .delete method on the client instead. This means instead of "
            "`my_client.partial_bid_matrix_information_with_scenarios.delete(my_ids)` please use `my_client.delete(my_ids)`."
            "The motivation is that all delete methods are the same, and having one delete method per API "
            " class encourages users to delete items in small batches, which is inefficient.",
            UserWarning,
            stacklevel=2,
        )
        return self._delete(external_id, space)

    @overload
    def retrieve(
        self, external_id: str, space: str = DEFAULT_INSTANCE_SPACE
    ) -> PartialBidMatrixInformationWithScenarios | None: ...

    @overload
    def retrieve(
        self, external_id: SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> PartialBidMatrixInformationWithScenariosList: ...

    def retrieve(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> PartialBidMatrixInformationWithScenarios | PartialBidMatrixInformationWithScenariosList | None:
        """Retrieve one or more partial bid matrix information with scenarios by id(s).

        Args:
            external_id: External id or list of external ids of the partial bid matrix information with scenarios.
            space: The space where all the partial bid matrix information with scenarios are located.

        Returns:
            The requested partial bid matrix information with scenarios.

        Examples:

            Retrieve partial_bid_matrix_information_with_scenario by id:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> partial_bid_matrix_information_with_scenario = client.partial_bid_matrix_information_with_scenarios.retrieve("my_partial_bid_matrix_information_with_scenario")

        """
        return self._retrieve(
            external_id,
            space,
            retrieve_edges=True,
            edge_api_name_type_direction_view_id_penta=[
                (
                    self.alerts_edge,
                    "alerts",
                    dm.DirectRelationReference("power_ops_types", "calculationIssue"),
                    "outwards",
                    dm.ViewId("power_ops_core", "Alert", "1"),
                ),
                (
                    self.underlying_bid_matrices_edge,
                    "underlying_bid_matrices",
                    dm.DirectRelationReference("power_ops_types", "intermediateBidMatrix"),
                    "outwards",
                    dm.ViewId("power_ops_core", "BidMatrix", "1"),
                ),
                (
                    self.multi_scenario_input_edge,
                    "multi_scenario_input",
                    dm.DirectRelationReference("power_ops_types", "calculationIssue"),
                    "outwards",
                    dm.ViewId("power_ops_core", "PriceProduction", "1"),
                ),
            ],
        )

    def search(
        self,
        query: str,
        properties: (
            PartialBidMatrixInformationWithScenariosTextFields
            | Sequence[PartialBidMatrixInformationWithScenariosTextFields]
            | None
        ) = None,
        state: str | list[str] | None = None,
        state_prefix: str | None = None,
        power_asset: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_resource_cost: float | None = None,
        max_resource_cost: float | None = None,
        partial_bid_configuration: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> PartialBidMatrixInformationWithScenariosList:
        """Search partial bid matrix information with scenarios

        Args:
            query: The search query,
            properties: The property to search, if nothing is passed all text fields will be searched.
            state: The state to filter on.
            state_prefix: The prefix of the state to filter on.
            power_asset: The power asset to filter on.
            min_resource_cost: The minimum value of the resource cost to filter on.
            max_resource_cost: The maximum value of the resource cost to filter on.
            partial_bid_configuration: The partial bid configuration to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of partial bid matrix information with scenarios to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Search results partial bid matrix information with scenarios matching the query.

        Examples:

           Search for 'my_partial_bid_matrix_information_with_scenario' in all text properties:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> partial_bid_matrix_information_with_scenarios = client.partial_bid_matrix_information_with_scenarios.search('my_partial_bid_matrix_information_with_scenario')

        """
        filter_ = _create_partial_bid_matrix_information_with_scenario_filter(
            self._view_id,
            state,
            state_prefix,
            power_asset,
            min_resource_cost,
            max_resource_cost,
            partial_bid_configuration,
            external_id_prefix,
            space,
            filter,
        )
        return self._search(
            self._view_id,
            query,
            _PARTIALBIDMATRIXINFORMATIONWITHSCENARIOS_PROPERTIES_BY_FIELD,
            properties,
            filter_,
            limit,
        )

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: (
            PartialBidMatrixInformationWithScenariosFields
            | Sequence[PartialBidMatrixInformationWithScenariosFields]
            | None
        ) = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: (
            PartialBidMatrixInformationWithScenariosTextFields
            | Sequence[PartialBidMatrixInformationWithScenariosTextFields]
            | None
        ) = None,
        state: str | list[str] | None = None,
        state_prefix: str | None = None,
        power_asset: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_resource_cost: float | None = None,
        max_resource_cost: float | None = None,
        partial_bid_configuration: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]: ...

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: (
            PartialBidMatrixInformationWithScenariosFields
            | Sequence[PartialBidMatrixInformationWithScenariosFields]
            | None
        ) = None,
        group_by: (
            PartialBidMatrixInformationWithScenariosFields | Sequence[PartialBidMatrixInformationWithScenariosFields]
        ) = None,
        query: str | None = None,
        search_properties: (
            PartialBidMatrixInformationWithScenariosTextFields
            | Sequence[PartialBidMatrixInformationWithScenariosTextFields]
            | None
        ) = None,
        state: str | list[str] | None = None,
        state_prefix: str | None = None,
        power_asset: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_resource_cost: float | None = None,
        max_resource_cost: float | None = None,
        partial_bid_configuration: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList: ...

    def aggregate(
        self,
        aggregate: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: (
            PartialBidMatrixInformationWithScenariosFields
            | Sequence[PartialBidMatrixInformationWithScenariosFields]
            | None
        ) = None,
        group_by: (
            PartialBidMatrixInformationWithScenariosFields
            | Sequence[PartialBidMatrixInformationWithScenariosFields]
            | None
        ) = None,
        query: str | None = None,
        search_property: (
            PartialBidMatrixInformationWithScenariosTextFields
            | Sequence[PartialBidMatrixInformationWithScenariosTextFields]
            | None
        ) = None,
        state: str | list[str] | None = None,
        state_prefix: str | None = None,
        power_asset: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_resource_cost: float | None = None,
        max_resource_cost: float | None = None,
        partial_bid_configuration: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        """Aggregate data across partial bid matrix information with scenarios

        Args:
            aggregate: The aggregation to perform.
            property: The property to perform aggregation on.
            group_by: The property to group by when doing the aggregation.
            query: The query to search for in the text field.
            search_property: The text field to search in.
            state: The state to filter on.
            state_prefix: The prefix of the state to filter on.
            power_asset: The power asset to filter on.
            min_resource_cost: The minimum value of the resource cost to filter on.
            max_resource_cost: The maximum value of the resource cost to filter on.
            partial_bid_configuration: The partial bid configuration to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of partial bid matrix information with scenarios to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Aggregation results.

        Examples:

            Count partial bid matrix information with scenarios in space `my_space`:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> result = client.partial_bid_matrix_information_with_scenarios.aggregate("count", space="my_space")

        """

        filter_ = _create_partial_bid_matrix_information_with_scenario_filter(
            self._view_id,
            state,
            state_prefix,
            power_asset,
            min_resource_cost,
            max_resource_cost,
            partial_bid_configuration,
            external_id_prefix,
            space,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _PARTIALBIDMATRIXINFORMATIONWITHSCENARIOS_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: PartialBidMatrixInformationWithScenariosFields,
        interval: float,
        query: str | None = None,
        search_property: (
            PartialBidMatrixInformationWithScenariosTextFields
            | Sequence[PartialBidMatrixInformationWithScenariosTextFields]
            | None
        ) = None,
        state: str | list[str] | None = None,
        state_prefix: str | None = None,
        power_asset: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_resource_cost: float | None = None,
        max_resource_cost: float | None = None,
        partial_bid_configuration: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        """Produces histograms for partial bid matrix information with scenarios

        Args:
            property: The property to use as the value in the histogram.
            interval: The interval to use for the histogram bins.
            query: The query to search for in the text field.
            search_property: The text field to search in.
            state: The state to filter on.
            state_prefix: The prefix of the state to filter on.
            power_asset: The power asset to filter on.
            min_resource_cost: The minimum value of the resource cost to filter on.
            max_resource_cost: The maximum value of the resource cost to filter on.
            partial_bid_configuration: The partial bid configuration to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of partial bid matrix information with scenarios to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Bucketed histogram results.

        """
        filter_ = _create_partial_bid_matrix_information_with_scenario_filter(
            self._view_id,
            state,
            state_prefix,
            power_asset,
            min_resource_cost,
            max_resource_cost,
            partial_bid_configuration,
            external_id_prefix,
            space,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _PARTIALBIDMATRIXINFORMATIONWITHSCENARIOS_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        state: str | list[str] | None = None,
        state_prefix: str | None = None,
        power_asset: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_resource_cost: float | None = None,
        max_resource_cost: float | None = None,
        partial_bid_configuration: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        retrieve_edges: bool = True,
    ) -> PartialBidMatrixInformationWithScenariosList:
        """List/filter partial bid matrix information with scenarios

        Args:
            state: The state to filter on.
            state_prefix: The prefix of the state to filter on.
            power_asset: The power asset to filter on.
            min_resource_cost: The minimum value of the resource cost to filter on.
            max_resource_cost: The maximum value of the resource cost to filter on.
            partial_bid_configuration: The partial bid configuration to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of partial bid matrix information with scenarios to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.
            retrieve_edges: Whether to retrieve `alerts`, `underlying_bid_matrices` or `multi_scenario_input` external ids for the partial bid matrix information with scenarios. Defaults to True.

        Returns:
            List of requested partial bid matrix information with scenarios

        Examples:

            List partial bid matrix information with scenarios and limit to 5:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> partial_bid_matrix_information_with_scenarios = client.partial_bid_matrix_information_with_scenarios.list(limit=5)

        """
        filter_ = _create_partial_bid_matrix_information_with_scenario_filter(
            self._view_id,
            state,
            state_prefix,
            power_asset,
            min_resource_cost,
            max_resource_cost,
            partial_bid_configuration,
            external_id_prefix,
            space,
            filter,
        )

        return self._list(
            limit=limit,
            filter=filter_,
            retrieve_edges=retrieve_edges,
            edge_api_name_type_direction_view_id_penta=[
                (
                    self.alerts_edge,
                    "alerts",
                    dm.DirectRelationReference("power_ops_types", "calculationIssue"),
                    "outwards",
                    dm.ViewId("power_ops_core", "Alert", "1"),
                ),
                (
                    self.underlying_bid_matrices_edge,
                    "underlying_bid_matrices",
                    dm.DirectRelationReference("power_ops_types", "intermediateBidMatrix"),
                    "outwards",
                    dm.ViewId("power_ops_core", "BidMatrix", "1"),
                ),
                (
                    self.multi_scenario_input_edge,
                    "multi_scenario_input",
                    dm.DirectRelationReference("power_ops_types", "calculationIssue"),
                    "outwards",
                    dm.ViewId("power_ops_core", "PriceProduction", "1"),
                ),
            ],
        )



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING

from cognite.client import data_modeling as dm, CogniteClient

from cognite.powerops.client._generated.v1.data_classes import (
    DomainModelCore,
    PlantWaterValueBased,
)
from cognite.powerops.client._generated.v1.data_classes._generator import (
    Generator,
    _create_generator_filter,
)
from ._core import DEFAULT_QUERY_LIMIT, QueryBuilder, QueryStep, QueryAPI, T_DomainModelList, _create_edge_filter

if TYPE_CHECKING:
    from .generator_query import GeneratorQueryAPI


class PlantWaterValueBasedQueryAPI(QueryAPI[T_DomainModelList]):
    def __init__(
        self,
        client: CogniteClient,
        builder: QueryBuilder[T_DomainModelList],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId],
        filter_: dm.filters.Filter | None = None,
        limit: int = DEFAULT_QUERY_LIMIT,
    ):
        super().__init__(client, builder, view_by_read_class)

        self._builder.append(
            QueryStep(
                name=self._builder.next_name("plant_water_value_based"),
                expression=dm.query.NodeResultSetExpression(
                    from_=self._builder[-1].name if self._builder else None,
                    filter=filter_,
                ),
                select=dm.query.Select(
                    [dm.query.SourceSelector(self._view_by_read_class[PlantWaterValueBased], ["*"])]
                ),
                result_cls=PlantWaterValueBased,
                max_retrieve_limit=limit,
            )
        )

    def generators(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        asset_type: str | list[str] | None = None,
        asset_type_prefix: str | None = None,
        min_production_min: float | None = None,
        max_production_min: float | None = None,
        min_penstock_number: int | None = None,
        max_penstock_number: int | None = None,
        min_start_stop_cost: float | None = None,
        max_start_stop_cost: float | None = None,
        generator_efficiency_curve: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        external_id_prefix_edge: str | None = None,
        space_edge: str | list[str] | None = None,
        filter: dm.Filter | None = None,
        limit: int | None = DEFAULT_QUERY_LIMIT,
    ) -> GeneratorQueryAPI[T_DomainModelList]:
        """Query along the generator edges of the plant water value based.

        Args:
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            display_name: The display name to filter on.
            display_name_prefix: The prefix of the display name to filter on.
            min_ordering: The minimum value of the ordering to filter on.
            max_ordering: The maximum value of the ordering to filter on.
            asset_type: The asset type to filter on.
            asset_type_prefix: The prefix of the asset type to filter on.
            min_production_min: The minimum value of the production min to filter on.
            max_production_min: The maximum value of the production min to filter on.
            min_penstock_number: The minimum value of the penstock number to filter on.
            max_penstock_number: The maximum value of the penstock number to filter on.
            min_start_stop_cost: The minimum value of the start stop cost to filter on.
            max_start_stop_cost: The maximum value of the start stop cost to filter on.
            generator_efficiency_curve: The generator efficiency curve to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            external_id_prefix_edge: The prefix of the external ID to filter on.
            space_edge: The space to filter on.
            filter: (Advanced) Filter applied to node. If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.
            limit: Maximum number of generator edges to return. Defaults to 3. Set to -1, float("inf") or None
                to return all items.

        Returns:
            GeneratorQueryAPI: The query API for the generator.
        """
        from .generator_query import GeneratorQueryAPI

        from_ = self._builder[-1].name
        edge_filter = _create_edge_filter(
            dm.DirectRelationReference("power_ops_types", "isSubAssetOf"),
            external_id_prefix=external_id_prefix_edge,
            space=space_edge,
        )
        self._builder.append(
            QueryStep(
                name=self._builder.next_name("generators"),
                expression=dm.query.EdgeResultSetExpression(
                    filter=edge_filter,
                    from_=from_,
                    direction="outwards",
                ),
                select=dm.query.Select(),
                max_retrieve_limit=limit,
            )
        )

        view_id = self._view_by_read_class[Generator]
        has_data = dm.filters.HasData(views=[view_id])
        node_filer = _create_generator_filter(
            view_id,
            name,
            name_prefix,
            display_name,
            display_name_prefix,
            min_ordering,
            max_ordering,
            asset_type,
            asset_type_prefix,
            min_production_min,
            max_production_min,
            min_penstock_number,
            max_penstock_number,
            min_start_stop_cost,
            max_start_stop_cost,
            generator_efficiency_curve,
            external_id_prefix,
            space,
            (filter and dm.filters.And(filter, has_data)) or has_data,
        )
        return GeneratorQueryAPI(self._client, self._builder, self._view_by_read_class, node_filer, limit)

    def query(
        self,
    ) -> T_DomainModelList:
        """Execute query and return the result.

        Returns:
            The list of the source nodes of the query.

        """
        return self._query()



"""
# How to do logging in this project

## Local development
1. Call `configure_debug_logging()` once, right after config is loaded.
2. Use vanilla Python logging:
   ```
   import logging
   logger = logging.getLogger(__name__)
   ```

## Production
1. Just use vanilla Python logging.
"""

import logging
import sys
from typing import Union

from loguru import logger

__all__ = [
    "configure_debug_logging",
]


LoggingLevelT = Union[int, str]


# https://github.com/Delgan/loguru#entirely-compatible-with-standard-logging
class InterceptHandler(logging.Handler):
    def emit(self, record):
        # Get corresponding Loguru level if it exists.
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message.
        frame, depth = sys._getframe(6), 6
        while frame and frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())


def configure_debug_logging(level: LoggingLevelT) -> None:
    logging.basicConfig(handlers=[InterceptHandler()], level=level, force=True)



from __future__ import annotations

import logging
from collections.abc import Sequence as SequenceType
from pathlib import Path

import yaml
from typing_extensions import Self

from cognite.powerops.client.shop.shop_file_reference import SHOPFileReference, SHOPFileType

logger = logging.getLogger(__name__)


class SHOPCase:
    r"""
    Wrapper around YAML file for SHOP, describing a case.

    Main features:
     * Values can be retrieved or set / changed like with a dict.
     * Can hold references to additional files ("extra files").
        * These are just externalID values of files in CDF, their content is not accessed here.
     * Loading from string or file (with `SHOPCase(filepath=path_to_yaml)`).
     * Saving to yaml (with `case.save_yaml(path_to_file)`).

    Examples:
      * load a case from string
          >>> case = SHOPCase('''
          ... foo:
          ...   bar1: 11
          ...   bar2: 22
          ... ''')
          >>> case.data
          {'foo': {'bar1': 11, 'bar2': 22}}

      * edit a value and show updated data
          >>> case.data["foo"]["bar2"] = 202
          >>> case.yaml
          'foo:\n  bar1: 11\n  bar2: 202\n'

      * load from and save to a file:
          case = SHOPCase(file_path="path/to/my_case.yaml")
          case[...] = ...  # edit data
          case.save_yaml("path/to/same_or_different.yaml")
    """

    def __init__(
        self,
        data: str = "",
        *,
        file_path: str = "",
        shop_files: SequenceType[SHOPFileReference] = (),
        watercourse: str = "",
        shop_version: str = "",
    ) -> None:
        if data and file_path:
            raise ValueError("Cannot specify both data and file_path")
        self.data = {}
        self.excess_yaml_parts: list[str] = []
        if data:
            self.load_case_data(data)
        elif file_path:
            self.load_case_file(file_path)
        self._shop_files: list[SHOPFileReference] = list(shop_files)
        self.watercourse = watercourse
        self.shop_version = shop_version

    def load_case_data(self, data: str) -> None:
        self.data = {}
        self.excess_yaml_parts = []
        try:
            if yaml_docs := list(yaml.safe_load_all(data)):
                self.data = yaml_docs[0]
                self._handle_excess_yaml_parts(yaml_docs[1:])
        except yaml.YAMLError as e:
            raise ValueError("Could not parse case data") from e

    def load_case_file(self, file_path: str, encoding: str = "utf-8") -> Self:
        with Path(file_path).open(encoding=encoding) as file:
            self.load_case_data(file.read())
        return self

    @property
    def shop_files(self) -> list[SHOPFileReference]:
        return self._shop_files.copy()

    def add_shop_file(self, external_id: str, shop_file_type: SHOPFileType = SHOPFileType.ASCII):
        logger.info(f"adding shop file from CDF: {external_id!r}, {shop_file_type=!r}")
        shop_file_reference = SHOPFileReference(external_id=external_id, file_type=shop_file_type)
        self.shop_files.append(shop_file_reference)

    @property
    def yaml(self) -> str:
        case_data = yaml.dump(self.data, sort_keys=False)
        if not case_data.endswith("\n"):
            case_data += "\n"
        return "---\n".join([case_data, *self.excess_yaml_parts])

    def save_yaml(self, path: str, encoding: str = "utf-8") -> None:
        logger.info(f"Saving case file to: {path}")
        with Path(path).open("w", encoding=encoding) as output_file:
            output_file.write(self.yaml)

    def _handle_excess_yaml_parts(self, excess_yaml_parts: list[str]) -> None:
        """
        If `Case.__init__` gets a yaml string which has multiple documents (separated by "---"),
        only the first document is parsed and set to `self.data`. Any subsequent documents are stored
        as "excess_yaml_parts". They are not part of `self.data`, but are kept and appended to the output.
        """
        if excess_yaml_parts:
            logger.warning(
                f"Case file contains {len(excess_yaml_parts) + 1} YAML documents. Only the first document is parsed,"
                f' additional documents will be passed to SHOP verbatim as "extra files".'
            )
        for yaml_doc in excess_yaml_parts:
            self.excess_yaml_parts.append(yaml.dump(yaml_doc, sort_keys=False))



from __future__ import annotations

from pathlib import Path

from cognite.client import ClientConfig, CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.credentials import OAuthClientCredentials

from ._api.case import CaseAPI
from ._api.commands_config import CommandsConfigAPI
from ._api.file_ref import FileRefAPI
from ._api.mapping import MappingAPI
from ._api.model_template import ModelTemplateAPI
from ._api.processing_log import ProcessingLogAPI
from ._api.scenario import ScenarioAPI
from ._api.transformation import TransformationAPI


class CogShop1Client:
    """
    CogShop1Client

    Generated with:
        pygen = 0.27.1
        cognite-sdk = 6.37.0
        pydantic = 2.4.2

    Data Model:
        space: cogShop
        externalId: CogShop
        version: 3
    """

    def __init__(self, config_or_client: CogniteClient | ClientConfig):
        if isinstance(config_or_client, CogniteClient):
            client = config_or_client
        elif isinstance(config_or_client, ClientConfig):
            client = CogniteClient(config_or_client)
        else:
            raise ValueError(f"Expected CogniteClient or ClientConfig, got {type(config_or_client)}")
        # The client name is used for aggregated logging of Pygen Usage
        client.config.client_name = "CognitePygen:0.27.1"

        self.case = CaseAPI(client, dm.ViewId("cogShop", "Case", "f532a799a59e1b"))
        self.commands_config = CommandsConfigAPI(client, dm.ViewId("cogShop", "CommandsConfig", "1bada7c1b95263"))
        self.file_ref = FileRefAPI(client, dm.ViewId("cogShop", "FileRef", "7c508a562e95ca"))
        self.mapping = MappingAPI(client, dm.ViewId("cogShop", "Mapping", "fc518056c47cad"))
        self.model_template = ModelTemplateAPI(client, dm.ViewId("cogShop", "ModelTemplate", "7864e537bac36e"))
        self.processing_log = ProcessingLogAPI(client, dm.ViewId("cogShop", "ProcessingLog", "c4fa72c3c45945"))
        self.scenario = ScenarioAPI(client, dm.ViewId("cogShop", "Scenario", "dcb646ab5384db"))
        self.transformation = TransformationAPI(client, dm.ViewId("cogShop", "Transformation", "917a6a16276e22"))

    @classmethod
    def azure_project(
        cls, tenant_id: str, client_id: str, client_secret: str, cdf_cluster: str, project: str
    ) -> CogShop1Client:
        credentials = OAuthClientCredentials.default_for_azure_ad(tenant_id, client_id, client_secret, cdf_cluster)
        config = ClientConfig.default(project, cdf_cluster, credentials)

        return cls(config)

    @classmethod
    def from_toml(cls, file_path: Path | str, section: str | None = "cognite") -> CogShop1Client:
        import toml

        toml_content = toml.load(file_path)
        if section is not None:
            try:
                toml_content = toml_content[section]
            except KeyError as e:
                raise ValueError(f"Could not find section '{section}' in {file_path}") from e

        return cls.azure_project(**toml_content)



from __future__ import annotations

import itertools
from typing import Any, Callable, ClassVar, Literal

from cognite.client import data_modeling as dm
from pydantic import Field, ValidationError

from cognite.powerops.client._generated.assets import data_classes as assets
from cognite.powerops.client.powerops_client import PowerOpsClient
from cognite.powerops.resync.models.base import Model
from cognite.powerops.utils.serialization import get_pydantic_annotation


class PowerAssetModelDM(Model):
    price_areas: list[assets.PriceAreaApply] = Field(default_factory=list)
    watercourses: list[assets.WatercourseApply] = Field(default_factory=list)
    plants: list[assets.PlantApply] = Field(default_factory=list)
    generators: list[assets.GeneratorApply] = Field(default_factory=list)
    reservoirs: list[assets.ReservoirApply] = Field(default_factory=list)
    turbine_curves: list[assets.TurbineEfficiencyCurveApply] = Field(default_factory=list)

    _views_by_write_class: ClassVar[dict[assets.DomainModelWrite, dm.ViewId]] = {
        assets.BidMethodApply: dm.ViewId("power-ops-shared", "BidMethod", "1"),
        assets.GeneratorApply: dm.ViewId("power-ops-assets", "Generator", "1"),
        assets.GeneratorEfficiencyCurveApply: dm.ViewId("power-ops-assets", "GeneratorEfficiencyCurve", "1"),
        assets.PlantApply: dm.ViewId("power-ops-assets", "Plant", "1"),
        assets.PriceAreaApply: dm.ViewId("power-ops-assets", "PriceArea", "1"),
        assets.ReservoirApply: dm.ViewId("power-ops-assets", "Reservoir", "1"),
        assets.TurbineEfficiencyCurveApply: dm.ViewId("power-ops-assets", "TurbineEfficiencyCurve", "1"),
        assets.WatercourseApply: dm.ViewId("power-ops-assets", "Watercourse", "1"),
    }

    def instances(self) -> dm.InstancesApply:
        resources = None
        cache = set()
        for item in itertools.chain(
            self.price_areas, self.generators, self.plants, self.reservoirs, self.watercourses, self.turbine_curves
        ):
            resource = item._to_instances_write(cache=cache, view_by_read_class=None)
            if resources is None:
                resources = resource
            else:
                resources.extend(resource)
        if resources is None:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))
        else:
            return dm.InstancesApply(nodes=resources.nodes, edges=resources.edges)

    def nodes(self) -> dm.NodeApplyList:
        return self.instances().nodes

    def edges(self) -> dm.EdgeApplyList:
        return self.instances().edges

    cdf_resources: ClassVar[dict[Callable, type]] = {
        **dict(Model.cdf_resources.items()),
        nodes: dm.NodeApply,
        edges: dm.EdgeApply,
    }

    @classmethod
    def from_cdf(cls, client: PowerOpsClient, data_set_external_id: str) -> PowerAssetModelDM:
        cdf = client.cdf
        # This takes advantage of the fact that every single edge in the model is of type "isSubAssetOf"
        is_type = dm.filters.Equals(["edge", "type"], {"externalId": "isSubAssetOf", "space": "power-ops-types"})
        edges = cdf.data_modeling.instances.list("edge", limit=-1, filter=is_type)
        edges = dm.EdgeApplyList([e.as_apply() for e in edges])
        nodes = dm.NodeApplyList([])
        for view in cls._views_by_write_class.values():
            view_nodes = cdf.data_modeling.instances.list("node", limit=-1, sources=view)
            nodes.extend([n.as_apply() for n in view_nodes])

        return cls._load(nodes, edges, link="external_id")

    @classmethod
    def load_from_cdf_resources(
        cls, data: dict[str, Any], link: Literal["external_id", "object"] = "object"
    ) -> PowerAssetModelDM:
        if not data:
            return cls()
        nodes = dm.NodeApplyList.load(data.get("nodes", []))
        edges = dm.EdgeApplyList.load(data.get("edges", []))
        return cls._load(nodes, edges, link)

    @classmethod
    def _load(
        cls, nodes: dm.NodeApplyList, edges: dm.EdgeApplyList, link: Literal["external_id", "object"] = "object"
    ) -> PowerAssetModelDM:
        write_class_by_view = {view: write_class for write_class, view in cls._views_by_write_class.items()}

        nodes_by_id: dict[tuple[str, str], assets.DomainModelApply] = {}
        for node in nodes:
            write_class = write_class_by_view[node.sources[0].source]
            unpack_node = node.dump(camel_case=False)
            sources = unpack_node.pop("sources", [])
            for source in sources:
                # Todo: For a generic case you need a check on view is matching data class here
                for key, value in source["properties"].items():
                    if isinstance(value, dict) and "space" in value and "externalId" in value:
                        unpack_node[key] = value["externalId"]
                    else:
                        unpack_node[key] = value

            field_names = {
                v for field_name, field in write_class.model_fields.items() for v in [field_name, field.alias] if v
            }
            try:
                domain_node = write_class(**{k: v for k, v in unpack_node.items() if k in field_names})
            except ValidationError:
                # My guess is that you get nodes with missing properties because of soft delete. So we just skip them.
                continue
            nodes_by_id[domain_node.as_tuple_id()] = domain_node

        for edge in edges:
            start_node = nodes_by_id.get((edge.start_node.space, edge.start_node.external_id))
            end_node = nodes_by_id.get((edge.end_node.space, edge.end_node.external_id))
            if start_node is None or end_node is None:
                continue
            for field_name, field in start_node.model_fields.items():
                annotation, _ = get_pydantic_annotation(field.annotation, type(start_node))
                if isinstance(end_node, annotation):
                    new_value = end_node if link == "object" else end_node.external_id
                    if (current_value := getattr(start_node, field_name)) is None:
                        setattr(start_node, field_name, [new_value])
                    else:
                        current_value.append(new_value)
                    break

        # Solve direct relations
        for node in nodes_by_id.values():
            for field_name, field in node.model_fields.items():
                annotation, _ = get_pydantic_annotation(field.annotation, type(node))
                if issubclass(annotation, assets.DomainModelWrite):
                    current_value = getattr(node, field_name)
                    if current_value is None:
                        continue
                    elif isinstance(current_value, (str, dm.NodeId)):
                        if isinstance(current_value, dm.NodeId):
                            space = current_value.space
                            external_id = current_value.external_id
                        else:
                            space = node.space
                            external_id = current_value
                        if (domain_node := nodes_by_id.get((space, external_id))) is not None:
                            setattr(node, field_name, domain_node)

        args = {}
        for field_name, field in cls.model_fields.items():
            annotation, _ = get_pydantic_annotation(field.annotation, cls)
            args[field_name] = [node for node in nodes_by_id.values() if isinstance(node, annotation)]
        return cls(**args)

    def standardize(self) -> None:
        return None



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

if TYPE_CHECKING:
    from ._file_ref import FileRefApply
    from ._mapping import MappingApply

__all__ = [
    "ModelTemplate",
    "ModelTemplateApply",
    "ModelTemplateList",
    "ModelTemplateApplyList",
    "ModelTemplateFields",
    "ModelTemplateTextFields",
]


ModelTemplateTextFields = Literal["version", "shop_version", "watercourse", "source"]
ModelTemplateFields = Literal["version", "shop_version", "watercourse", "source"]

_MODELTEMPLATE_PROPERTIES_BY_FIELD = {
    "version": "version",
    "shop_version": "shopVersion",
    "watercourse": "watercourse",
    "source": "source",
}


class ModelTemplate(DomainModel):
    space: str = "cogShop"
    version: Optional[str] = None
    shop_version: Optional[str] = Field(None, alias="shopVersion")
    watercourse: Optional[str] = None
    model: Optional[str] = None
    source: Optional[str] = None
    base_mappings: Optional[list[str]] = Field(None, alias="baseMappings")

    def as_apply(self) -> ModelTemplateApply:
        return ModelTemplateApply(
            external_id=self.external_id,
            version=self.version,
            shop_version=self.shop_version,
            watercourse=self.watercourse,
            model=self.model,
            source=self.source,
            base_mappings=self.base_mappings,
        )


class ModelTemplateApply(DomainModelApply):
    space: str = "cogShop"
    version: str
    shop_version: str = Field(alias="shopVersion")
    watercourse: str
    model: Union[FileRefApply, str, None] = Field(None, repr=False)
    source: Optional[str] = None
    base_mappings: Union[list[MappingApply], list[str], None] = Field(default=None, repr=False, alias="baseMappings")

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.version is not None:
            properties["version"] = self.version
        if self.shop_version is not None:
            properties["shopVersion"] = self.shop_version
        if self.watercourse is not None:
            properties["watercourse"] = self.watercourse
        if self.model is not None:
            properties["model"] = {
                "space": "cogShop",
                "externalId": self.model if isinstance(self.model, str) else self.model.external_id,
            }
        if self.source is not None:
            properties["source"] = self.source
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("cogShop", "ModelTemplate"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        for base_mapping in self.base_mappings or []:
            edge = self._create_base_mapping_edge(base_mapping)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(base_mapping, DomainModelApply):
                instances = base_mapping._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        if isinstance(self.model, DomainModelApply):
            instances = self.model._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))

    def _create_base_mapping_edge(self, base_mapping: Union[str, MappingApply]) -> dm.EdgeApply:
        if isinstance(base_mapping, str):
            end_node_ext_id = base_mapping
        elif isinstance(base_mapping, DomainModelApply):
            end_node_ext_id = base_mapping.external_id
        else:
            raise TypeError(f"Expected str or MappingApply, got {type(base_mapping)}")

        return dm.EdgeApply(
            space="cogShop",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("cogShop", "ModelTemplate.baseMappings"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("cogShop", end_node_ext_id),
        )


class ModelTemplateList(TypeList[ModelTemplate]):
    _NODE = ModelTemplate

    def as_apply(self) -> ModelTemplateApplyList:
        return ModelTemplateApplyList([node.as_apply() for node in self.data])


class ModelTemplateApplyList(TypeApplyList[ModelTemplateApply]):
    _NODE = ModelTemplateApply



from __future__ import annotations

import datetime
from collections.abc import Sequence
from typing import overload
import warnings

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.v1.data_classes._core import DEFAULT_INSTANCE_SPACE
from cognite.powerops.client._generated.v1.data_classes import (
    DomainModelCore,
    DomainModelWrite,
    ResourcesWriteResult,
    ShopCase,
    ShopCaseWrite,
    ShopCaseFields,
    ShopCaseList,
    ShopCaseWriteList,
)
from cognite.powerops.client._generated.v1.data_classes._shop_case import (
    _SHOPCASE_PROPERTIES_BY_FIELD,
    _create_shop_case_filter,
)
from ._core import (
    DEFAULT_LIMIT_READ,
    DEFAULT_QUERY_LIMIT,
    Aggregations,
    NodeAPI,
    SequenceNotStr,
    QueryStep,
    QueryBuilder,
)
from .shop_case_shop_files import ShopCaseShopFilesAPI
from .shop_case_query import ShopCaseQueryAPI


class ShopCaseAPI(NodeAPI[ShopCase, ShopCaseWrite, ShopCaseList]):
    def __init__(self, client: CogniteClient, view_by_read_class: dict[type[DomainModelCore], dm.ViewId]):
        view_id = view_by_read_class[ShopCase]
        super().__init__(
            client=client,
            sources=view_id,
            class_type=ShopCase,
            class_list=ShopCaseList,
            class_write_list=ShopCaseWriteList,
            view_by_read_class=view_by_read_class,
        )
        self._view_id = view_id
        self.shop_files_edge = ShopCaseShopFilesAPI(client)

    def __call__(
        self,
        scenario: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_start_time: datetime.datetime | None = None,
        max_start_time: datetime.datetime | None = None,
        min_end_time: datetime.datetime | None = None,
        max_end_time: datetime.datetime | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_QUERY_LIMIT,
        filter: dm.Filter | None = None,
    ) -> ShopCaseQueryAPI[ShopCaseList]:
        """Query starting at shop cases.

        Args:
            scenario: The scenario to filter on.
            min_start_time: The minimum value of the start time to filter on.
            max_start_time: The maximum value of the start time to filter on.
            min_end_time: The minimum value of the end time to filter on.
            max_end_time: The maximum value of the end time to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of shop cases to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            A query API for shop cases.

        """
        has_data = dm.filters.HasData(views=[self._view_id])
        filter_ = _create_shop_case_filter(
            self._view_id,
            scenario,
            min_start_time,
            max_start_time,
            min_end_time,
            max_end_time,
            external_id_prefix,
            space,
            (filter and dm.filters.And(filter, has_data)) or has_data,
        )
        builder = QueryBuilder(ShopCaseList)
        return ShopCaseQueryAPI(self._client, builder, self._view_by_read_class, filter_, limit)

    def apply(
        self,
        shop_case: ShopCaseWrite | Sequence[ShopCaseWrite],
        replace: bool = False,
        write_none: bool = False,
    ) -> ResourcesWriteResult:
        """Add or update (upsert) shop cases.

        Note: This method iterates through all nodes and timeseries linked to shop_case and creates them including the edges
        between the nodes. For example, if any of `shop_files` are set, then these
        nodes as well as any nodes linked to them, and all the edges linking these nodes will be created.

        Args:
            shop_case: Shop case or sequence of shop cases to upsert.
            replace (bool): How do we behave when a property value exists? Do we replace all matching and existing values with the supplied values (true)?
                Or should we merge in new values for properties together with the existing values (false)? Note: This setting applies for all nodes or edges specified in the ingestion call.
            write_none (bool): This method, will by default, skip properties that are set to None. However, if you want to set properties to None,
                you can set this parameter to True. Note this only applies to properties that are nullable.
        Returns:
            Created instance(s), i.e., nodes, edges, and time series.

        Examples:

            Create a new shop_case:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> from cognite.powerops.client._generated.v1.data_classes import ShopCaseWrite
                >>> client = PowerOpsModelsV1Client()
                >>> shop_case = ShopCaseWrite(external_id="my_shop_case", ...)
                >>> result = client.shop_case.apply(shop_case)

        """
        warnings.warn(
            "The .apply method is deprecated and will be removed in v1.0. "
            "Please use the .upsert method on the client instead. This means instead of "
            "`my_client.shop_case.apply(my_items)` please use `my_client.upsert(my_items)`."
            "The motivation is that all apply methods are the same, and having one apply method per API "
            " class encourages users to create items in small batches, which is inefficient."
            "In addition, .upsert method is more descriptive of what the method does.",
            UserWarning,
            stacklevel=2,
        )
        return self._apply(shop_case, replace, write_none)

    def delete(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> dm.InstancesDeleteResult:
        """Delete one or more shop case.

        Args:
            external_id: External id of the shop case to delete.
            space: The space where all the shop case are located.

        Returns:
            The instance(s), i.e., nodes and edges which has been deleted. Empty list if nothing was deleted.

        Examples:

            Delete shop_case by id:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> client.shop_case.delete("my_shop_case")
        """
        warnings.warn(
            "The .delete method is deprecated and will be removed in v1.0. "
            "Please use the .delete method on the client instead. This means instead of "
            "`my_client.shop_case.delete(my_ids)` please use `my_client.delete(my_ids)`."
            "The motivation is that all delete methods are the same, and having one delete method per API "
            " class encourages users to delete items in small batches, which is inefficient.",
            UserWarning,
            stacklevel=2,
        )
        return self._delete(external_id, space)

    @overload
    def retrieve(self, external_id: str, space: str = DEFAULT_INSTANCE_SPACE) -> ShopCase | None: ...

    @overload
    def retrieve(self, external_id: SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE) -> ShopCaseList: ...

    def retrieve(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> ShopCase | ShopCaseList | None:
        """Retrieve one or more shop cases by id(s).

        Args:
            external_id: External id or list of external ids of the shop cases.
            space: The space where all the shop cases are located.

        Returns:
            The requested shop cases.

        Examples:

            Retrieve shop_case by id:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> shop_case = client.shop_case.retrieve("my_shop_case")

        """
        return self._retrieve(
            external_id,
            space,
            retrieve_edges=True,
            edge_api_name_type_direction_view_id_penta=[
                (
                    self.shop_files_edge,
                    "shop_files",
                    dm.DirectRelationReference("power_ops_types", "ShopCase.shopFiles"),
                    "outwards",
                    dm.ViewId("power_ops_core", "ShopFile", "1"),
                ),
            ],
        )

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: ShopCaseFields | Sequence[ShopCaseFields] | None = None,
        group_by: None = None,
        scenario: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_start_time: datetime.datetime | None = None,
        max_start_time: datetime.datetime | None = None,
        min_end_time: datetime.datetime | None = None,
        max_end_time: datetime.datetime | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]: ...

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: ShopCaseFields | Sequence[ShopCaseFields] | None = None,
        group_by: ShopCaseFields | Sequence[ShopCaseFields] = None,
        scenario: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_start_time: datetime.datetime | None = None,
        max_start_time: datetime.datetime | None = None,
        min_end_time: datetime.datetime | None = None,
        max_end_time: datetime.datetime | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList: ...

    def aggregate(
        self,
        aggregate: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: ShopCaseFields | Sequence[ShopCaseFields] | None = None,
        group_by: ShopCaseFields | Sequence[ShopCaseFields] | None = None,
        scenario: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_start_time: datetime.datetime | None = None,
        max_start_time: datetime.datetime | None = None,
        min_end_time: datetime.datetime | None = None,
        max_end_time: datetime.datetime | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        """Aggregate data across shop cases

        Args:
            aggregate: The aggregation to perform.
            property: The property to perform aggregation on.
            group_by: The property to group by when doing the aggregation.
            scenario: The scenario to filter on.
            min_start_time: The minimum value of the start time to filter on.
            max_start_time: The maximum value of the start time to filter on.
            min_end_time: The minimum value of the end time to filter on.
            max_end_time: The maximum value of the end time to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of shop cases to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Aggregation results.

        Examples:

            Count shop cases in space `my_space`:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> result = client.shop_case.aggregate("count", space="my_space")

        """

        filter_ = _create_shop_case_filter(
            self._view_id,
            scenario,
            min_start_time,
            max_start_time,
            min_end_time,
            max_end_time,
            external_id_prefix,
            space,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _SHOPCASE_PROPERTIES_BY_FIELD,
            property,
            group_by,
            None,
            None,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: ShopCaseFields,
        interval: float,
        scenario: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_start_time: datetime.datetime | None = None,
        max_start_time: datetime.datetime | None = None,
        min_end_time: datetime.datetime | None = None,
        max_end_time: datetime.datetime | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        """Produces histograms for shop cases

        Args:
            property: The property to use as the value in the histogram.
            interval: The interval to use for the histogram bins.
            scenario: The scenario to filter on.
            min_start_time: The minimum value of the start time to filter on.
            max_start_time: The maximum value of the start time to filter on.
            min_end_time: The minimum value of the end time to filter on.
            max_end_time: The maximum value of the end time to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of shop cases to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Bucketed histogram results.

        """
        filter_ = _create_shop_case_filter(
            self._view_id,
            scenario,
            min_start_time,
            max_start_time,
            min_end_time,
            max_end_time,
            external_id_prefix,
            space,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _SHOPCASE_PROPERTIES_BY_FIELD,
            None,
            None,
            limit,
            filter_,
        )

    def list(
        self,
        scenario: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_start_time: datetime.datetime | None = None,
        max_start_time: datetime.datetime | None = None,
        min_end_time: datetime.datetime | None = None,
        max_end_time: datetime.datetime | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        retrieve_edges: bool = True,
    ) -> ShopCaseList:
        """List/filter shop cases

        Args:
            scenario: The scenario to filter on.
            min_start_time: The minimum value of the start time to filter on.
            max_start_time: The maximum value of the start time to filter on.
            min_end_time: The minimum value of the end time to filter on.
            max_end_time: The maximum value of the end time to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of shop cases to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.
            retrieve_edges: Whether to retrieve `shop_files` external ids for the shop cases. Defaults to True.

        Returns:
            List of requested shop cases

        Examples:

            List shop cases and limit to 5:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> shop_cases = client.shop_case.list(limit=5)

        """
        filter_ = _create_shop_case_filter(
            self._view_id,
            scenario,
            min_start_time,
            max_start_time,
            min_end_time,
            max_end_time,
            external_id_prefix,
            space,
            filter,
        )

        return self._list(
            limit=limit,
            filter=filter_,
            retrieve_edges=retrieve_edges,
            edge_api_name_type_direction_view_id_penta=[
                (
                    self.shop_files_edge,
                    "shop_files",
                    dm.DirectRelationReference("power_ops_types", "ShopCase.shopFiles"),
                    "outwards",
                    dm.ViewId("power_ops_core", "ShopFile", "1"),
                ),
            ],
        )
